# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
CLI program for storing pacakges in CKAN

The program uses the Root.Distributions in the source package to locate packages to link into a CKAN record.

"""

import shutil
from os import getenv, remove, listdir
from os.path import join, basename, exists, isdir, splitext

from metapack.package import *
from metapack.util import ensure_dir

from .core import MetapackCliMemo as _MetapackCliMemo
from .core import prt
import json

downloader = Downloader()

class MetapackCliMemo(_MetapackCliMemo):

    def __init__(self, args, downloader):
        super().__init__(args, downloader)
        self


def install_file(subparsers):

    parser = subparsers.add_parser(
        'file',
        help='Install packages to a filesystem'
    )

    parser.set_defaults(run_command=run_file_install)

    parser.add_argument('-l', '--list', default=False, action='store_true',
                             help="List installed packages")

    parser.add_argument('metatabfile', nargs='?', default='./',
                        help='Path to a Metatab file. Optional; if not specified, use metatab.csv for package specification ')

    if getenv('METAPACK_DATA'):
       d_args = '?'
       defaults_to=f"Defaults to METAPACK_DATA env var, '{getenv('METAPACK_DATA')}'"
    else:
        d_args = 1
        defaults_to = 'May also be set with the METAPACK_DATA env var'

    parser.add_argument('directory', nargs=d_args, default=getenv('METAPACK_DATA'),
                        help='Directory to which to install files. '+defaults_to)

def run_file_install(args):

    if args.list:
        index_path = join(args.directory, 'index.json')

        if exists(index_path):
            with open(index_path) as f:
                packages = json.load(f)
                for n in packages.keys():
                    if not n.startswith('_'):
                        prt(n)
        return

    try:
        args.directory = args.directory.pop(0) # Its a list
    except AttributeError:
        pass # its a string

    m = MetapackCliMemo(args, downloader=downloader)

    copy_packages(m)

def find_packages(name, pkg_dir):
    """Locate pre-built packages in the _packages directory"""
    for c in (FileSystemPackageBuilder, ZipPackageBuilder, ExcelPackageBuilder):

        package_path, cache_path = c.make_package_path(pkg_dir, name)

        if package_path.exists():

            yield c.type_code, package_path, cache_path #c(package_path, pkg_dir)

def copy_packages(m):

    for ptype, purl, cache_path in find_packages(m.doc.get_value('Root.Name'), m.package_root):

        dest = join(m.args.directory, basename(purl.path))

        ensure_dir(m.args.directory)

        if ptype in ('xlsx', 'zip'):
            copy_file(m, purl.path, dest)

        elif ptype == 'fs':
            copy_dir(m, purl.path, dest)


def copy_file(m, source, dest):

    if exists(dest):
        remove(dest)

    prt(f"Copy file {source} to {dest} ")
    shutil.copy(source, dest)

    update_index(m.args.directory, join(dest))

def copy_dir(m, source, dest):

    if exists(dest):
        shutil.rmtree(dest)

    prt(f"Copy dir {source} to {dest}")
    shutil.copytree(source, dest)

    update_index(m.args.directory, dest)

def update_index(d, package_path, suffix=''):

    index_path = join(d,'index.json')

    def add_package_to_index(p, packages, suffix=suffix):

        def add_suffix(name):
            if not suffix:
                return name
            else:
                return name + '-' + suffix

        pkg = open_package(p)

        packages[add_suffix(pkg.get_value('Root.Identifier'))] = p # identifier

        packages[add_suffix(pkg._generate_identity_name())] = p # Fully qualified name

        nv_name = add_suffix((pkg._generate_identity_name(mod_version=None)))

        if not '_versions' in packages:
            packages['_versions'] = {}

        if not nv_name in packages['_versions']:
            packages['_versions'][nv_name] = {}

        packages['_versions'][nv_name][pkg.get_value('Root.Version')] = p

        # Find latest version
        versions = set()
        for ver in packages['_versions'][nv_name].keys():
            try:
                versions.add(int(ver))
            except ValueError:
                # Only doing integer versions name. Need to figure out
                # how to handle Semantic Versions, later.
                pass

        if versions:
            latest = max(versions)

            packages[nv_name] = packages['_versions'][nv_name][str(latest)]

            packages[nv_name+'-latest'] = packages['_versions'][nv_name][str(latest)]

        # Remove the unziped zip package. FIXME It really should just pull the
        # metadata file without unzipping the whole package.

        if exists(p + '_d'):
            shutil.rmtree(p + '_d')

    if exists(index_path):
        with open(index_path) as f:
            packages = json.load(f)

        add_package_to_index(package_path, packages, suffix=suffix)

        with open(index_path,'w') as f:
            json.dump(packages, f, indent=4)

    else:

        packages = {}

        def yield_packages(d):

            for e in listdir(d):
                path = join(d, e)
                bn, ext = splitext(path)
                if isdir(path):
                    if exists(join(path, 'metadata.csv')):
                        yield join(path, 'metadata.csv')
                elif ext in ('.xls', '.xlsx', '.zip'):
                    yield path

        for p in yield_packages(d):
            add_package_to_index(p, packages, suffix=None)

        ensure_dir(join(d,'source'))

        for p in yield_packages(join(d,'source')):
            add_package_to_index(p, packages, suffix='source')

        with open(index_path,'w') as f:
            json.dump(packages, f, indent=4)







# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
CLI program for storing pacakges in CKAN
"""

from os import getcwd, getenv
from os.path import basename

from botocore.exceptions import  NoCredentialsError


from metapack import MetapackDoc, MetapackUrl
from metapack import MetapackPackageUrl
from metapack.cli.core import prt, err, make_s3_package, PACKAGE_PREFIX
from metapack.package import *
from metapack.package.s3 import S3Bucket
from metapack.util import datetime_now
from metatab import DEFAULT_METATAB_FILE
from rowgenerators import parse_app_url
from rowgenerators.util import clean_cache
from rowgenerators.util import fs_join as join

class MetapackCliMemo(object):

    def __init__(self, args):
        from appurl import get_cache
        self.cwd = getcwd()

        self.args = args

        self.downloader = Downloader()

        self.cache = self.downloader.cache

        self.mtfile_arg = self.args.metatabfile if self.args.metatabfile else join(self.cwd, DEFAULT_METATAB_FILE)

        self.mtfile_url = MetapackUrl(self.mtfile_arg, downloader=self.downloader)

        self.resource = self.mtfile_url.fragment

        self.package_url = self.mtfile_url.package_url
        self.mt_file = self.mtfile_url.metadata_url

        self.package_root = self.package_url.join(PACKAGE_PREFIX)

        if not self.args.s3:
            doc = MetapackDoc(self.mt_file)
            self.args.s3 = doc['Root'].find_first_value('Root.S3')

        self.s3_url = parse_app_url(self.args.s3)

        if not self.s3_url.scheme == 's3':
            self.s3_url = parse_app_url("s3://{}".format(self.args.s3))

        self.doc = MetapackDoc(self.mt_file)

        access_value = self.doc.find_first_value('Root.Access')

        self.acl = 'private' if access_value == 'private' else 'public-read'

        self.bucket = S3Bucket(self.s3_url, acl=self.acl , profile=self.args.profile)


def metas3(subparsers):

    parser = subparsers.add_parser(
        's3',
        help='Create packages and store them in s3 buckets',
    )

    parser.set_defaults(run_command=run_s3)

    parser.add_argument('-p', '--profile', help="Name of a BOTO or AWS credentails profile", required=False)

    parser.add_argument('-s', '--s3', help="URL to S3 where packages will be stored", required=False)

    parser.add_argument('-C', '--credentials', help="Show S3 Credentials and exit. "
                                                    "Eval this string to setup credentials in other shells.",
                        action='store_true', default=False)

    parser.add_argument('metatabfile', nargs='?', help='Path to a Metatab file')


def find_packages(name, pkg_dir):
    """Locate pre-built packages in the _packages directory"""
    for c in (FileSystemPackageBuilder, ZipPackageBuilder, ExcelPackageBuilder):

        package_path, cache_path = c.make_package_path(pkg_dir, name)

        if package_path.exists():

            yield c.type_code, package_path, cache_path #c(package_path, pkg_dir)

def run_s3(args):

    m = MetapackCliMemo(args)

    if m.args.credentials:
        show_credentials(m.args.profile)
        exit(0)

    dist_urls = upload(m)

    if dist_urls:
        prt("Synchronized these Package Urls")
        prt("-------------------------------")
        for au in dist_urls:
            prt(au)
        prt("-------------------------------")

    else:
        prt("Did not find any packages to upload")


    set_distributions(m, dist_urls)

def set_distributions(m, dist_urls):

    for t in m.doc.find('Root.Distribution'):
        m.doc.remove_term(t)

    for au in dist_urls:
        m.doc['Root'].new_term('Root.Distribution', au)

    m.doc.write_csv()


def upload(m):

    dist_urls = []

    fs_p = None

    for ptype, purl, cache_path in find_packages(m.doc.get_value('Root.Name'), m.package_root):
        au = m.bucket.access_url(cache_path)

        if ptype in ('xlsx', 'zip'):
            with open(purl.path, mode='rb') as f:
                m.bucket.write(f.read(), basename(purl.path), m.acl)
                prt("Added {} distribution: {} ".format(ptype, au))
                dist_urls.append(au)

        elif ptype == 'fs':

            env = {}
            skip_if_exist = False

            try:
                s3_package_root = MetapackPackageUrl(str(m.s3_url), downloader=m.downloader)

                fs_p, fs_url, created = make_s3_package(purl.metadata_url, s3_package_root, m.cache, env, m.acl, skip_if_exist)
            except NoCredentialsError:
                print(getenv('AWS_SECRET_ACCESS_KEY'))
                err("Failed to find boto credentials for S3. "
                    "See http://boto3.readthedocs.io/en/latest/guide/configuration.html ")

            # A crappy hack. make_s3_package should return the correct url
            if fs_p:
                if m.acl == 'private':
                    au = fs_p.private_access_url.inner
                else:
                    au = fs_p.public_access_url.inner

                dist_urls.append(au)

    m.doc['Root'].get_or_new_term('Root.Issued').value = datetime_now()

    if dist_urls:

        # Create the CSV package, with links into the
        if fs_p:
            u = MetapackUrl(fs_p.access_url, downloader=m.downloader)

            resource_root = u.dirname().as_type(MetapackPackageUrl)

            p = CsvPackageBuilder(u, m.package_root, resource_root)

            for au in dist_urls:
                p.doc['Root'].new_term('Root.Distribution', au)

            csv_url = p.save()

            with open(csv_url.path, mode='rb') as f:
                m.bucket.write(f.read(), csv_url.target_file, m.acl)

            dist_urls.append(m.bucket.access_url(p.cache_path))
        else:
            # If this happens, then no packages were created, because an FS package
            # is always built first
            prt("Not creating CSV package; no FS package was uploaded")

    return dist_urls



def show_credentials(profile):
    import boto3

    session = boto3.Session(profile_name=profile)

    if profile:
        cred_line = " 'eval $(metasync -C -p {} )'".format(profile)
    else:
        cred_line = " 'eval $(metasync -C)'"

    prt("export AWS_ACCESS_KEY_ID={} ".format(session.get_credentials().access_key))
    prt("export AWS_SECRET_ACCESS_KEY={}".format(session.get_credentials().secret_key))
    prt("# Run {} to configure credentials in a shell".format(cred_line))


def run_docker(m):
    """Re-run the metasync command in docker. """

    import botocore.session
    from subprocess import Popen, PIPE, STDOUT

    raise NotImplementedError("No longer have access to raw_args")

    session = botocore.session.get_session()

    args = ['docker', 'run', '--rm', '-t', '-i',
            '-eAWS_ACCESS_KEY_ID={}'.format(session.get_credentials().access_key),
            '-eAWS_SECRET_ACCESS_KEY={}'.format(session.get_credentials().secret_key),
            'civicknowledge/metatab',
            'metasync']

    for a in ('-D', '--docker'):
        try:
            m.raw_args.remove(a)
        except ValueError:
            pass

    args.extend(m.raw_args[1:])

    if m.args.verbose:
        prt("Running Docker Command: ", ' '.join(args))
    else:
        prt("Running In Docker")

    process = Popen(args, stdout=PIPE, stderr=STDOUT)
    with process.stdout:
        for line in iter(process.stdout.readline, b''):
            prt(line.decode('ascii'), end='')

    exitcode = process.wait()  # 0 means success

    exit(exitcode)




# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
CLI program for storing pacakges in CKAN

The program uses the Root.Distributions in the source package to locate packages to link into a CKAN record.

"""

import shutil
from os import getenv, remove, listdir
from os.path import join, basename, exists, isdir, splitext

from metapack.package import *
from metapack.util import ensure_dir, slugify

from .core import MetapackCliMemo as _MetapackCliMemo
from .core import prt, err
import json
from .install_file import update_index

downloader = Downloader()

class MetapackCliMemo(_MetapackCliMemo):

    def __init__(self, args, downloader):
        super().__init__(args, downloader)

def install_git(subparsers):

    parser = subparsers.add_parser(
        'git',
        help='Install git repos as source directories'
    )

    parser.set_defaults(run_command=run_git_install)

    if getenv('METAPACK_DATA'):
       d_args = '?'
       defaults_to=f"Defaults to METAPACK_DATA env var, '{getenv('METAPACK_DATA')}'"
    else:
        d_args = 1
        defaults_to = 'May also be set with the METAPACK_DATA env var'

    parser.add_argument('-d', '--directory', nargs=d_args, default=getenv('METAPACK_DATA'),
                        help='Directory to which to install files. '+defaults_to)

    parser.add_argument('giturl', nargs=1, help='URL to a git repo')


def run_git_install(args):
    try:
        from git import RemoteProgress, Repo, GitCommandError, InvalidGitRepositoryError
    except ImportError:
        err("This command requires the git python module. Run 'pip install gitpython'")

    class MyProgressPrinter(RemoteProgress):
        def update(self, op_code, cur_count, max_count=None, message=''):
            print(op_code, cur_count, max_count, cur_count / (max_count or 100.0), message or "NO MESSAGE")

    if isinstance(args.giturl, list):
        giturl = args.giturl[0]
    else:
        giturl = args.giturl

    target = join(args.directory, 'source', slugify(giturl))

    ensure_dir(target)

    try:
        repo = Repo(target)
        repo.remotes.origin.pull()
        prt(f"Pulling to existing repo, '{target} ")
    except InvalidGitRepositoryError:
        repo = Repo.clone_from(giturl, target)
        prt(f"Cloning to new repo, '{target} ")

    update_index(args.directory, target, suffix='source')




import unittest
from rowgenerators import parse_app_url, match_url_classes, WebUrl, FileUrl, ZipUrl, CsvFileUrl

from metapack import MetapackDoc
from metapack.appurl import MetapackUrl, MetapackResourceUrl, MetapackDocumentUrl
from rowgenerators import get_generator
from csv import DictReader
from metapack.test.support import test_data
from os.path import exists


class TestUrls(unittest.TestCase):
    """Test Metapack AppUrls and Row Generators"""

    def test_metapack_urls(self):

        groups = {}

        with open(test_data('mpurls.csv')) as f:
            for l in DictReader(f):

                u = parse_app_url(l['in_url'])

                self.assertEqual(l['url_class'], u.__class__.__name__)
                self.assertEqual(l['url'], str(u))
                self.assertEqual(l['package_url'], str(u.package_url))

                # The second instance in each group is a resource url for the the
                # metadata url of the first instance.
                if (l['url_class'] == 'MetapackDocumentUrl'):

                    self.assertNotIn(l['group'], groups)
                    self.assertEqual(str(u), str(u.metadata_url))
                else:
                    self.assertIn(l['group'], groups)
                    self.assertEqual(str(groups[l['group']]), str(u.metadata_url))

                groups[l['group']] = u

                self.assertEqual(l['metadata_url'], str(u.metadata_url))

                r = u.get_resource()
                self.assertTrue(r.inner.exists())

                t = r.get_target()
                try:
                    self.assertTrue(t.inner.exists(), (type(r), t, t.inner))
                except AssertionError:
                    t = r.get_target()

                # Check that the generator for the metadata gets the right number of rows
                self.assertEqual(50, len(list(u.metadata_url.generator)))

                self.assertEqual('example.com-simple_example-2017-us-1', (u.doc.find_first_value('Root.Name')))

    def test_metapack_url(self):

        # Zip

        us = 'metapack+http://library.metatab.org/example.com-simple_example-2017-us-1.zip'
        ud = parse_app_url(us)

        self.assertEqual('metapack+http://library.metatab.org/example.com-simple_example-2017-us-1.zip#metadata.csv',
                         str(ud))
        self.assertIsInstance(ud, MetapackDocumentUrl)
        print("!!!", type(ud.inner))
        import sys
        print("!!!", sys.modules[type(ud.inner).__module__].__file__)
        self.assertIsInstance(ud.inner, WebUrl)

        r = ud.get_resource()

        self.assertIsInstance(r, MetapackDocumentUrl)
        self.assertIsInstance(r.inner, ZipUrl)

        t = r.get_target()
        self.assertIsInstance(t, CsvFileUrl)

        g = get_generator(t)
        self.assertEquals(50, len(list(g)))

        self.assertEquals('example.com-simple_example-2017-us-1', (ud.doc.find_first_value('Root.Name')))

        # Excel

        us = 'metapack+http://library.metatab.org/example.com-simple_example-2017-us-1.xlsx'
        ud = parse_app_url(us)

        self.assertEqual('metapack+http://library.metatab.org/example.com-simple_example-2017-us-1.xlsx#meta',
                         str(ud))
        self.assertIsInstance(ud, MetapackDocumentUrl)
        self.assertEquals('example.com-simple_example-2017-us-1', (ud.doc.find_first_value('Root.Name')))

        # Filesystem
        us = 'metapack+http://library.metatab.org/example.com-simple_example-2017-us-1/metadata.csv'
        ud = parse_app_url(us)

        self.assertEqual('metapack+http://library.metatab.org/example.com-simple_example-2017-us-1/metadata.csv',
                         str(ud))
        self.assertIsInstance(ud, MetapackDocumentUrl)
        self.assertEquals('example.com-simple_example-2017-us-1', (ud.doc.find_first_value('Root.Name')))

        # Filesystem
        us = 'metapack+http://library.metatab.org/example.com-simple_example-2017-us-1/'
        ud = parse_app_url(us)

        self.assertEqual('metapack+http://library.metatab.org/example.com-simple_example-2017-us-1/metadata.csv',
                         str(ud))
        self.assertIsInstance(ud, MetapackDocumentUrl)
        self.assertEquals('example.com-simple_example-2017-us-1', (ud.doc.find_first_value('Root.Name')))

        self.assertTrue(str(ud.get_resource()).endswith(
            'library.metatab.org/example.com-simple_example-2017-us-1/metadata.csv'))

        self.assertTrue(str(ud.get_resource().get_target()).endswith(
            'library.metatab.org/example.com-simple_example-2017-us-1/metadata.csv'))

        # -----
        # Resource Urls

        us = 'metapack+http://library.metatab.org.s3.amazonaws.com/example.com-simple_example-2017-us-1/metadata.csv#random_names'

        ur = parse_app_url(us)
        self.assertEqual('metapack', ur.proto)
        self.assertIsInstance(ur, MetapackResourceUrl)

    def test_inner(self):
        u_s = 'metapack+http://public.source.civicknowledge.com.s3.amazonaws.com/example.com/geo/Parks_SD.zip#encoding=utf8'

        u = parse_app_url(u_s)

        self.assertIsInstance(u, MetapackDocumentUrl)
        self.assertIsInstance(u.inner, WebUrl)

        r = u.get_resource()
        self.assertIsInstance(r, MetapackDocumentUrl)
        self.assertIsInstance(r.inner, ZipUrl)

    def test_fs_resource(self):

        us = 'metapack+http://library.metatab.org/example.com-simple_example-2017-us-1/#random_names'

        u = parse_app_url(us)

        self.assertIsInstance(u, MetapackResourceUrl, type(u))

        self.assertEquals('metapack+http://library.metatab.org/example.com-simple_example-2017-us-1/metadata.csv',
                          str(u.metadata_url))

        self.assertEquals('metapack+http://library.metatab.org/example.com-simple_example-2017-us-1/',
                          str(u.package_url))

    def test_xlsx_parse(self):

        ru = parse_app_url('/foo/bar/bax.xlsx',
                           fragment='fragment',
                           )

        # print (repr(ru))

    def test_metatab_resource(self):

        u_str = 'metapack+http://library.metatab.org/ffiec.gov-cra_aggregate_smb-orig-1.csv#sb_agg_loan_orig'

        u = parse_app_url(u_str)

        print(type(u))

        print(u.resource)

    def test_metatab_resource_zip(self):

        us = 'metapack+http://library.metatab.org/example.com-simple_example-2017-us-1.zip#random-names'

        u = parse_app_url(us)

        doc = u.metadata_url.doc

        r = doc.resource(u.target_file)

        self.assertEqual(101, len(list(r)))

    def test_metatab_resource_xlsx(self):

        us = 'metapack+http://library.metatab.org/example.com-simple_example-2017-us-1.xlsx#random-names'

        u = parse_app_url(us)
        self.assertIsInstance(u, MetapackResourceUrl)
        self.assertEqual('metapack+http://library.metatab.org/example.com-simple_example-2017-us-1.xlsx#random-names',
                         str(u))
        self.assertEqual('random-names', u.fragment[0])
        self.assertEqual('random-names', u.target_file)

        doc = u.doc

        r = doc.resource(u.target_file)
        ru = r.resolved_url
        # print(r.value)
        # print (ru)
        rur = ru.get_resource()
        # print (rur)
        rurt = rur.get_target()
        # print (rurt)

        self.assertEqual(101, len(list(r)))

    def test_metatab_resource_csv(self):

        us = 'metapack+http://library.metatab.org/example.com-simple_example-2017-us-1.csv#random-names'

        u = parse_app_url(us)
        doc = u.metadata_url.doc

        r = doc.resource(u.target_file)

        self.assertEqual(101, len(list(r)))

    def test_metatab_resource_fs(self):

        us = 'metapack+http://library.metatab.org/example.com-simple_example-2017-us-1/metadata.csv#random-names'

        u = parse_app_url(us)
        doc = u.metadata_url.doc

        r = u.resource
        ru = r.resolved_url

        # print(ru.inner)
        # print(ru.get_resource())

        return

    def test_metatab_resource_zip(self):
        from metapack.appurl import MetapackResourceUrl

        us = 'metapack+http://s3.amazonaws.com/library.metatab.org/ipums.org-income_homevalue-5.zip#income_homeval'

        u = parse_app_url(us)

        self.assertIsInstance(u, MetapackResourceUrl)


    def test_metatab_resource_fs_pkg(self):
        from metapack.appurl import MetapackDocumentUrl

        us = 'metapack+http://library.metatab.org/example.com-simple_example-2017-us-1/metadata.csv'

        u = parse_app_url(us)

        self.assertIsInstance(u, MetapackDocumentUrl)


    def test_search_url(self):
        from metapack.appurl import SearchUrl
        from os import getenv
        from unittest import SkipTest

        us = 'search:bc.edu-creations-comments#comments'

        u = parse_app_url(us)

        if getenv('METAPACK_DATA') is None:
            raise SkipTest("Test requires the METAPACK_DATA envoironment variable to be set")

        search_func = SearchUrl.search_indexed_directory(getenv('METAPACK_DATA'))

        SearchUrl.register_search(search_func)

        self.assertEqual('metapack+file:/Volumes/Storage/Data/metapack/bc.edu-creations-comments-3.zip#comments',
                          str(u.get_resource()))



if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*
# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
Exporter to convert a notebook into a Metatab package.
"""
import re
import logging
import sys
import copy
import io
import json
import metapack
from metapack.exc import MetapackError
from metapack.jupyter.exc import NotebookError
# from metapack.jupyter.markdown import MarkdownExporter
from nbconvert.exporters.markdown import MarkdownExporter
import nbformat
from nbconvert.exporters import Exporter
from nbconvert.exporters.html import HTMLExporter
from nbconvert.exporters.pdf import PDFExporter
from nbconvert.exporters.latex import LatexExporter
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert.preprocessors import ExtractOutputPreprocessor
from nbconvert.preprocessors.execute import CellExecutionError
from os import getcwd
from os.path import dirname, join, abspath, basename
from traitlets import List
from traitlets.config import Unicode, Config, Dict
from .preprocessors import (AddEpilog, ExtractInlineMetatabDoc, RemoveMetatab,
                            ExtractFinalMetatabDoc, ExtractMetatabTerms, ExtractLibDirs)
from datetime import datetime
from traitlets import default

from metatab.util import ensure_dir


def write_files(self, resources):
    self.log.info('Base dir: {}'.format(self.output_dir))

    for filename, data in resources.get('outputs', {}).items():
        dest = join(self.output_dir, filename)

        ensure_dir(dest)

        with io.open(dest, 'wb') as f:
            f.write(data)
            self.log.info("Wrote '{}' ".format(filename))


class MetatabExporter(Exporter):
    template_path = List(['.']).tag(config=True, affects_environment=True)

    output_dir = Unicode(help='Output directory').tag(config=True)
    notebook_dir = Unicode(help='CWD in which notebook will be executed').tag(config=True)
    package_dir = Unicode(help='Directory in which to store generated package').tag(config=True)
    package_name = Unicode(help='Name of package to generate. Defaults to the Metatab Root.Name').tag(config=True)

    def __init__(self, config=None, **kw):
        # import pdb; pdb.set_trace();
        super().__init__(config, **kw)

        self.log = kw.get('log', logging.getLogger(self.__class__.__name__))

    def from_file(self, file_stream, resources=None, **kw):
        return super().from_file(file_stream, resources, **kw)

    def from_filename(self, filename, resources=None, **kw):
        if not self.notebook_dir:
            self.notebook_dir = dirname(abspath(filename))

        return super().from_filename(filename, resources, **kw)


class DocumentationExporter(MetatabExporter):
    """Exports multiple forms of documentation"""

    metadata = Dict(help='Extra metadata, added to the \'metatab\' key', default_value={}).tag(config=True)

    @property
    def default_config(self):
        import metapack.jupyter.templates

        c = Config()

        c.TemplateExporter.template_path = [dirname(metapack.jupyter.templates.__file__)]

        c.HTMLExporter.preprocessors = [
            'metapack.jupyter.preprocessors.NoShowInput',
            'metapack.jupyter.preprocessors.RemoveMetatab',
            'metapack.jupyter.preprocessors.HtmlBib'
        ]

        c.HTMLExporter.exclude_input_prompt = True
        c.HTMLExporter.exclude_output_prompt = True

        c.MarkdownExporter.preprocessors = ['metapack.jupyter.preprocessors.RemoveMagics']

        c.PDFExporter.preprocessors = [
            # 'metapack.jupyter.preprocessors.NoShowInput',
            'metapack.jupyter.preprocessors.RemoveMetatab',
            'metapack.jupyter.preprocessors.LatexBib',
            'metapack.jupyter.preprocessors.MoveTitleDescription'
        ]

        c.PDFExporter.exclude_input_prompt = True
        # Excluding the output prompt also excludes the output tables.
        # .PDFExporter.exclude_output_prompt = True

        c.merge(super(DocumentationExporter, self).default_config)
        return c

    def from_notebook_node(self, nb, resources=None, **kw):

        nb_copy = copy.deepcopy(nb)

        nb_copy['metadata']['metatab'] = self.metadata

        # get the Normal HTML output:
        output, resources = HTMLExporter(config=self.config).from_notebook_node(nb_copy)

        resources['unique_key'] = 'notebook'

        # Get all of the image resources
        nb_copy, resources = self.extract_resources(nb_copy, resources)

        # Add resources for the html and markdown versionf of the notebook
        self.add_pdf(nb_copy, resources)
        self.add_markdown_doc(nb_copy, resources)
        self.add_html_doc(nb_copy, resources)
        self.add_basic_html_doc(nb_copy, resources)

        return output, resources

    def extract_resources(self, nb, resources):

        output_filename_template = "image_{cell_index}_{index}{extension}"

        return ExtractOutputPreprocessor(output_filename_template=output_filename_template) \
            .preprocess(nb, resources)

    def add_pdf(self, nb, resources):

        template_file = 'notebook.tplx'

        exp = PDFExporter(config=self.config, template_file=template_file)

        (body, _) = exp.from_notebook_node(nb)

        resources['outputs']['documentation.pdf'] = body

        exp = LatexExporter(config=self.config, template_file=template_file)

        (body, _) = exp.from_notebook_node(nb)

        resources['outputs']['documentation.latex'] = body.encode('utf-8')

    def add_basic_html_doc(self, nb, resources):
        html_exp = HTMLExporter(config=self.config, template_file='hide_input_html_basic.tpl')

        (html_basic_body, _) = html_exp.from_notebook_node(nb)

        resources['outputs']['html_basic_body.html'] = html_basic_body.encode('utf-8')

    def add_html_doc(self, nb, resources):
        html_exp = HTMLExporter(config=self.config, template_file='hide_input_html.tpl')

        (html_full_body, _) = html_exp.from_notebook_node(nb)

        resources['outputs']['documentation.html'] = html_full_body.encode('utf-8')

    def add_markdown_doc(self, nb, resources):

        exp = MarkdownExporter(config=self.config)
        (md_body, _) = exp.from_notebook_node(nb)

        resources['outputs']['documentation.md'] = md_body.encode('utf-8')

    def update_metatab(self, doc, resources):
        """Add documentation entries for resources"""

        if not 'Documentation' in doc:
            doc.new_section("Documentation")

        ds = doc['Documentation']

        # This is the main output from the HTML exporter, not a resource.
        ds.new_term('Root.Documentation', 'docs/notebook.html', name="notebook.html", title='Jupyter Notebook (HTML)')

        for name, data in resources.get('outputs', {}).items():

            if name == 'documentation.html':
                ds.new_term('Root.Documentation', 'docs/' + name, name=name, title='Primary Documentation (HTML)')

            elif name == 'html_basic_body.html':
                pass
            elif name.endswith('.html'):
                ds.new_term('Root.Documentation', 'docs/' + name, name=name, title='Documentation (HTML)')
            elif name.endswith('.md'):
                ds.new_term('Root.Documentation', 'docs/' + name, name=name, title='Documentation (Markdown)')
            elif name.endswith('.pdf'):
                ds.new_term('Root.Documentation', 'docs/' + name, name=name, title='Documentation (PDF)')
            elif name.endswith('.png'):
                ds.new_term('Root.Image', 'docs/' + name, name=name, title='Image for HTML Documentation')
            else:
                pass


class NotebookExecutor(MetatabExporter):
    """

    """

    file_extension = ''

    _preprocessors = List([

    ])

    extra_terms = []  # Terms set in the document

    lib_dirs = []  # Extra library directories

    doc = None

    @property
    def default_config(self):

        c = Config()

        c.ExecutePreprocessor.timeout = 600

        c.merge(super(NotebookExecutor, self).default_config)
        return c

    def get_package_dir_name(self, nb):
        """This is the name of the package we will be creating. """

        package_dir = self.package_dir

        if not package_dir:
            package_dir = getcwd()

        package_name = self.package_name

        if not package_name:
            doc = ExtractInlineMetatabDoc(package_url="metapack+file:" + package_dir).run(nb)

            if not doc:
                raise NotebookError("Notebook does not have an inline metatab doc")

            t = doc.find_first('Root.Name', section='Root')

            if not t:
                raise NotebookError("Inline Metatab doc doesnt have a Root.Name term")

            package_name = doc.as_version(None)

        return package_dir, package_name

    def get_output_dir(self, nb):
        """Open a notebook and determine the output directory from the name"""
        self.package_dir, self.package_name = self.get_package_dir_name(nb)

        return join(self.package_dir, self.package_name)

    def extract_terms(self, nb):
        """Extract some term values, usually set with tags or metadata"""
        emt = ExtractMetatabTerms()
        emt.preprocess(nb, {})

        return emt.terms

    def write_notebook(self, nb, path='/tmp/executed_notebook.ipynb'):
        with open(path, 'wt') as f:
            nbformat.write(nb, f)

        return path

    def from_notebook_node(self, nb, resources=None, **kw):
        """Create a Metatab package from a notebook node """

        nb_copy = copy.deepcopy(nb)

        # The the package name and directory, either from the inlined Metatab doc,
        # or from the config

        try:
            self.output_dir = self.get_output_dir(nb)
        except NotebookError as e:
            # Notebook probably lacks a metatab doc.
            self.log.fatal(e)
            sys.exit(1)

        assert self.output_dir

        resources = self._init_resources(resources)

        resources['outputs'] = {}

        if 'language' in nb['metadata']:
            resources['language'] = nb['metadata']['language'].lower()

        # Do any other configured preprocessing
        nb_copy, resources = self._preprocess(nb_copy, resources)

        # The Notebook can set some terms with tags
        self.extra_terms = self.extract_terms(nb_copy)

        # Clear the output before executing
        self.clear_output(nb_copy)

        nb_copy, resources = self.exec_notebook(nb_copy, resources, self.notebook_dir)

        eld = ExtractLibDirs()
        eld.preprocess(nb_copy, {})

        self.lib_dirs = eld.lib_dirs

        efm = ExtractFinalMetatabDoc()
        efm.preprocess(nb_copy, {})

        if not efm.doc:
            raise MetapackError("No metatab doc")

        self.doc = efm.doc

        for section, term, value in self.extra_terms:
            self.doc[section].get_or_new_term(term, value)

        nb, _ = RemoveMetatab().preprocess(nb, {})

        resources['outputs']['notebooks/{}.ipynb'.format(self.package_name)] = nbformat.writes(nb).encode('utf-8')

        return efm.doc.as_csv(), resources

    def clear_output(self, nb):

        from nbconvert.preprocessors import ClearOutputPreprocessor

        return ClearOutputPreprocessor().preprocess(nb, {})

    def exec_notebook(self, nb, resources, nb_dir):

        nb_copy, _ = AddEpilog(config=self.config, pkg_dir=self.output_dir).preprocess(nb, resources)

        resources['outputs']['notebooks/executed-source.ipynb'] = nbformat.writes(nb).encode('utf-8')

        exec_nb_path = self.write_notebook(nb_copy)

        try:
            ep = ExecutePreprocessor(config=self.config)

            nb, _ = ep.preprocess(nb, {'metadata': {'path': nb_dir}})
        except CellExecutionError as e:
            raise CellExecutionError("Errors executing noteboook. See output at {} for details.\n{}"
                                     .format(exec_nb_path, ''))

        return nb, resources


class HugoOutputExtractor(ExtractOutputPreprocessor):
    def preprocess(self, nb, resources):
        return super().preprocess(nb, resources)

    def preprocess(self, nb, resources):

        for index, cell in enumerate(nb.cells):
            nb.cells[index], resources = self.preprocess_cell(cell, resources, index)

        return nb, resources

    def preprocess_cell(self, cell, resources, cell_index):
        """Also extracts attachments"""
        from nbformat.notebooknode import NotebookNode

        attach_names = []

        # Just move the attachment into an output

        for k, attach in cell.get('attachments', {}).items():
            for mime_type in self.extract_output_types:
                if mime_type in attach:

                    if not 'outputs' in cell:
                        cell['outputs'] = []

                    o = NotebookNode({
                        'data': NotebookNode({mime_type: attach[mime_type]}),
                        'metadata': NotebookNode({
                            'filenames': {mime_type: k}  # Will get re-written
                        }),
                        'output_type': 'display_data'
                    })

                    cell['outputs'].append(o)

                    attach_names.append((mime_type, k))

        nb, resources = super().preprocess_cell(cell, resources, cell_index)

        output_names = list(resources.get('outputs', {}).keys())

        if attach_names:
            # We're going to assume that attachments are only on Markdown cells, and Markdown cells
            # can't generate output, so all of the outputs wee added.

            # reverse + zip matches the last len(attach_names) elements from output_names

            for output_name, (mimetype, an) in zip(reversed(output_names), reversed(attach_names)):
                # We'll post process to set the final output directory
                cell.source = re.sub('\(attachment:{}\)'.format(an),
                                     '(__IMGDIR__/{})'.format(output_name), cell.source)

        return nb, resources


class HugoExporter(MarkdownExporter):
    """ Export a python notebook to markdown, with frontmatter for Hugo.
    """

    hugo_dir = Unicode(help="Root of the Hugo directory").tag(config=True)

    section = Unicode(help="Hugo section in which to write the converted notebook").tag(config=True)

    @default('section')
    def _section_file_default(self):
        return 'notebooks'

    @property
    def default_config(self):
        import metapack.jupyter.templates

        c = Config({

        })

        c.TemplateExporter.template_path = [dirname(metapack.jupyter.templates.__file__)]
        c.TemplateExporter.template_file = 'markdown_hugo.tpl'

        c.MarkdownExporter.preprocessors = [
            'metapack.jupyter.preprocessors.OrganizeMetadata',
            HugoOutputExtractor
        ]

        c.merge(super(HugoExporter, self).default_config)

        c.ExtractOutputPreprocessor.enabled = False

        return c

    def get_creators(self, meta):

        for typ in ('wrangler', 'creator'):
            try:
                # Multiple authors
                for e in meta[typ]:
                    d = dict(e.items())
                    d['type'] = typ

                    yield d
            except AttributeError:
                # only one
                d = meta[typ]
                d['type'] = typ
                yield d
            except KeyError:
                pass

    def from_notebook_node(self, nb, resources=None, **kw):

        nb_copy = copy.deepcopy(nb)

        resources = self._init_resources(resources)

        if 'language' in nb['metadata']:
            resources['language'] = nb['metadata']['language'].lower()

        # Preprocess
        nb_copy, resources = self._preprocess(nb_copy, resources)

        # move over some more metadata
        if 'authors' not in nb_copy.metadata.frontmatter:
            nb_copy.metadata.frontmatter['authors'] = list(self.get_creators(nb_copy.metadata.metatab))

        # Other useful metadata
        if not 'date' in nb_copy.metadata.frontmatter:
            nb_copy.metadata.frontmatter['date'] = datetime.now().isoformat()

        resources.setdefault('raw_mimetypes', self.raw_mimetypes)
        resources['global_content_filter'] = {
            'include_code': not self.exclude_code_cell,
            'include_markdown': not self.exclude_markdown,
            'include_raw': not self.exclude_raw,
            'include_unknown': not self.exclude_unknown,
            'include_input': not self.exclude_input,
            'include_output': not self.exclude_output,
            'include_input_prompt': not self.exclude_input_prompt,
            'include_output_prompt': not self.exclude_output_prompt,
            'no_prompt': self.exclude_input_prompt and self.exclude_output_prompt,
        }

        slug = nb_copy.metadata.frontmatter.slug

        # Rebuild all of the image names
        for cell_index, cell in enumerate(nb_copy.cells):
            for output_index, out in enumerate(cell.get('outputs', [])):

                if 'metadata' in out:
                    for type_name, fn in list(out.metadata.get('filenames', {}).items()):
                        if fn in resources['outputs']:
                            html_path = join('img', slug, basename(fn))
                            file_path = join(self.hugo_dir, 'static', html_path)

                            resources['outputs'][file_path] = resources['outputs'][fn]
                            del resources['outputs'][fn]

                            # Can't put the '/' in the join() or it will be absolute

                            out.metadata.filenames[type_name] = '/' + html_path

        output = self.template.render(nb=nb_copy, resources=resources)

        section = nb_copy.metadata.frontmatter.get('section') or self.section

        # Don't know why this isn't being set from the config
        # resources['output_file_dir'] = self.config.NbConvertApp.output_base

        # Setting full path to subvert the join() in the file writer. I can't
        # figure out how to set the output directories from this function
        resources['unique_key'] = join(self.hugo_dir, 'content', section, slug)

        # Probably should be done with a postprocessor.
        output = re.sub(r'__IMGDIR__', join('/img', slug), output)

        return output, resources


class WordpressExporter(HTMLExporter):
    """ Export a python notebook to markdown, with frontmatter for Hugo.
    """

    staging_dir = Unicode(help="Root of the Hugo directory").tag(config=True)

    @property
    def default_config(self):
        import metapack.jupyter.templates

        c = Config({

        })

        c.TemplateExporter.template_path = [dirname(metapack.jupyter.templates.__file__)]
        c.TemplateExporter.template_file = 'html_wordpress.tpl'

        c.HTMLExporter.preprocessors = [
            'metapack.jupyter.preprocessors.OrganizeMetadata',
            HugoOutputExtractor
        ]

        c.merge(super(WordpressExporter, self).default_config)

        c.ExtractOutputPreprocessor.enabled = False

        return c

    def get_creators(self, meta):

        for typ in ('wrangler', 'creator'):
            try:
                # Multiple authors
                for e in meta[typ]:
                    d = dict(e.items())
                    d['type'] = typ

                    yield d
            except AttributeError:
                # only one
                d = meta[typ]
                d['type'] = typ
                yield d
            except KeyError:
                pass

    def from_notebook_node(self, nb, resources=None, **kw):
        from nbconvert.filters.highlight import Highlight2HTML
        import nbformat

        nb_copy = copy.deepcopy(nb)

        resources = self._init_resources(resources)

        if 'language' in nb['metadata']:
            resources['language'] = nb['metadata']['language'].lower()

        # Preprocess
        nb_copy, resources = self._preprocess(nb_copy, resources)

        # move over some more metadata
        if 'authors' not in nb_copy.metadata.frontmatter:
            nb_copy.metadata.frontmatter['authors'] = list(self.get_creators(nb_copy.metadata.metatab))

        # Other useful metadata
        if not 'date' in nb_copy.metadata.frontmatter:
            nb_copy.metadata.frontmatter['date'] = datetime.now().isoformat()

        resources.setdefault('raw_mimetypes', self.raw_mimetypes)

        resources['global_content_filter'] = {
            'include_code': not self.exclude_code_cell,
            'include_markdown': not self.exclude_markdown,
            'include_raw': not self.exclude_raw,
            'include_unknown': not self.exclude_unknown,
            'include_input': not self.exclude_input,
            'include_output': not self.exclude_output,
            'include_input_prompt': not self.exclude_input_prompt,
            'include_output_prompt': not self.exclude_output_prompt,
            'no_prompt': self.exclude_input_prompt and self.exclude_output_prompt,
        }

        langinfo = nb.metadata.get('language_info', {})
        lexer = langinfo.get('pygments_lexer', langinfo.get('name', None))
        self.register_filter('highlight_code',
                             Highlight2HTML(pygments_lexer=lexer, parent=self))

        def format_datetime(value, format='%a, %B %d'):
            from  dateutil.parser import parse
            return parse(value).strftime(format)

        self.register_filter('parsedatetime', format_datetime)

        slug = nb_copy.metadata.frontmatter.slug

        # Rebuild all of the image names
        for cell_index, cell in enumerate(nb_copy.cells):
            for output_index, out in enumerate(cell.get('outputs', [])):

                if 'metadata' in out:
                    for type_name, fn in list(out.metadata.get('filenames', {}).items()):
                        if fn in resources['outputs']:
                            html_path = join(slug, basename(fn))
                            file_path = join(self.staging_dir, html_path)

                            resources['outputs'][file_path] = resources['outputs'][fn]
                            del resources['outputs'][fn]

                            # Can't put the '/' in the join() or it will be absolute

                            out.metadata.filenames[type_name] = '/' + html_path

        output = self.template.render(nb=nb_copy, resources=resources)

        # Don't know why this isn't being set from the config
        # resources['output_file_dir'] = self.config.NbConvertApp.output_base

        # Setting full path to subvert the join() in the file writer. I can't
        # figure out how to set the output directories from this function
        resources['unique_key'] = join(self.staging_dir, slug)

        # Probably should be done with a postprocessor.
        output = re.sub(r'__IMGDIR__', '/' + slug, output)

        output = re.sub(r'<style>.*</style>', '', output, flags=re.MULTILINE | re.DOTALL)

        resources['outputs'][join(self.staging_dir, slug + '.json')] = \
            json.dumps(nb_copy.metadata, indent=4).encode('utf-8')

        resources['outputs'][join(self.staging_dir, slug + '.ipynb')] = nbformat.writes(nb_copy).encode('utf-8')

        return output, resources




#from metatab.cli.metatab import metatab
#from metatab.cli.metapack import metapack
#from metatab.cli.metakan import metakan
#from metatab.cli.metasync import metasync
#from metatab.cli.metaworld import metaworld


# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE
"""
Record objects for the Simple Data Package format.
"""

from rowgenerators import get_generator, parse_app_url
from .exc import *
from .doc import MetapackDoc, Resolver
from .package import open_package, Downloader
from .appurl import MetapackUrl, MetapackDocumentUrl, MetapackResourceUrl, MetapackPackageUrl
from .terms import Resource

from metapack.jupyter.magic import load_ipython_extension, unload_ipython_extension




import logging
import sys
from genericpath import exists
from itertools import islice
from os.path import join
from uuid import uuid4
from os import getenv
from metapack.appurl import SearchUrl

import six
from tableintuit import TypeIntuiter

from metapack import MetapackDoc, MetapackUrl
from metapack.package.csv import CsvPackageBuilder
from metapack.package.excel import ExcelPackageBuilder
from metapack.package.filesystem import FileSystemPackageBuilder
from metapack.package.s3 import S3PackageBuilder
from metapack.package.zip import ZipPackageBuilder
from metatab import  DEFAULT_METATAB_FILE
from metatab.util import make_metatab_file
from rowgenerators import SelectiveRowGenerator, parse_app_url, AppUrlError

logger = logging.getLogger('user')
logger_err = logging.getLogger('cli-errors')
debug_logger = logging.getLogger('debug')

def cli_init(log_level=logging.INFO):

    out_hdlr = logging.StreamHandler(sys.stdout)
    out_hdlr.setFormatter(logging.Formatter('%(message)s'))
    out_hdlr.setLevel(log_level)
    logger.addHandler(out_hdlr)
    logger.setLevel(log_level)

    out_hdlr = logging.StreamHandler(sys.stderr)
    out_hdlr.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))
    out_hdlr.setLevel(logging.WARN)
    logger_err.addHandler(out_hdlr)
    logger_err.setLevel(logging.WARN)

    if getenv('METAPACK_DATA'):
        try:
            search_func = SearchUrl.search_indexed_directory(getenv('METAPACK_DATA'))
            SearchUrl.register_search(search_func)
        except AppUrlError as e:
            warn("Failed to register package search function: "+str(e))

def prt(*args, **kwargs):
    logger.info(' '.join(str(e) for e in args),**kwargs)

def warn(*args, **kwargs):
    logger_err.warn(' '.join(str(e) for e in args),**kwargs)

def err(*args, **kwargs):
    logger_err.critical(' '.join(str(e) for e in args),**kwargs)
    sys.exit(1)


def metatab_info(cache):
   pass


def new_metatab_file(mt_file, template):
    template = template if template else 'metatab'

    if not exists(mt_file):
        doc = make_metatab_file(template)

        doc['Root']['Identifier'] = str(uuid4())

        doc.write_csv(mt_file)


def find_files(base_path, types):
    from os import walk
    from os.path import join, splitext

    for root, dirs, files in walk(base_path):
        if '_metapack' in root:
            continue

        for f in files:
            if f.startswith('_'):
                continue

            b, ext = splitext(f)
            if ext[1:] in types:
                yield join(root, f)


def get_lib_module_dict(doc):
    return doc.get_lib_module_dict()


def dump_resources(doc):
    for r in doc.resources():
        prt(r.name, r.resolved_url)


def dump_resource(doc, name, lines=None):
    import unicodecsv as csv
    import sys
    from itertools import islice
    from tabulate import tabulate
    from rowgenerators.rowpipe.exceptions import CasterExceptionError, TooManyCastingErrors

    r = doc.resource(name=name)

    if not r:
        err("Did not get resource for name '{}'".format(name))

    # WARNING! This code will not generate errors if line is set ( as for the -H
    # option because the errors are tansfered from the row pipe to the resource after the
    # iterator is exhausted

    gen = islice(r, 1, lines)

    def dump_errors(error_set):
        for col, errors in error_set.items():
            warn("Errors in casting column '{}' in resource '{}' ".format(col, r.name))
            for error in errors:
                warn("    ", error)


    try:
        if lines and lines <= 20:
            try:
                prt(tabulate(list(gen), list(r.headers)))
            except TooManyCastingErrors as e:
                dump_errors(e.errors)
                err(e)

        else:

            w = csv.writer(sys.stdout if six.PY2 else sys.stdout.buffer)

            if r.headers:
                w.writerow(r.headers)
            else:
                warn("No headers for resource '{}'; have schemas been generated? ".format(name))

            for row in gen:
                w.writerow(row)

    except CasterExceptionError as e:  # Really bad errors, not just casting problems.
        raise e
        err(e)
    except TooManyCastingErrors as e:
        dump_errors(e.errors)
        err(e)

    dump_errors(r.errors)



def dump_schema(doc, name):
    from tabulate import tabulate

    t = get_table(doc, name)

    rows = []
    header = 'name altname datatype description'.split()
    for c in t.children:
        cp = c.properties
        rows.append([cp.get(h) for h in header])

    prt(tabulate(rows, header))


def get_table(doc, name):
    t = doc.find_first('Root.Table', value=name)

    if not t:

        table_names = ["'" + t.value + "'" for t in doc.find('Root.Table')]

        if not table_names:
            table_names = ["<No Tables>"]

        err("Did not find schema for table name '{}' Tables are: {}"
            .format(name, " ".join(table_names)))

    return t

PACKAGE_PREFIX = '_packages'

def make_excel_package(file, package_root, cache, env, skip_if_exists):

    assert package_root

    p = ExcelPackageBuilder(file, package_root, callback=prt,  env=env)
    prt('Making Excel Package')

    if not p.exists() or not skip_if_exists:
        url = p.save()
        prt("Packaged saved to: {}".format(url))
        created = True
    elif p.exists():
        prt("Excel Package already exists")
        created = False
        url = p.package_path.path

    return p, MetapackUrl(url, downloader=package_root.downloader), created


def make_zip_package(file, package_root, cache, env, skip_if_exists):

    assert package_root


    p = ZipPackageBuilder(file, package_root, callback=prt,  env=env)
    prt('Making ZIP Package')
    if not p.exists() or not skip_if_exists:
        url = p.save()
        prt("Packaged saved to: {}".format(url))
        created = True
    elif p.exists():
        prt("ZIP Package already exists")
        created = False
        url = p.package_path.path

    return p, MetapackUrl(url, downloader=package_root.downloader), created


def make_filesystem_package(file, package_root, cache, env, skip_if_exists):

    assert package_root

    p = FileSystemPackageBuilder(file, package_root, callback=prt, env=env)

    if skip_if_exists is None:
        skip_if_exists = p.is_older_than_metatada()

    if not p.exists() or not skip_if_exists:
        prt('Making Filesystem Package ',
            '; existing package is older than metadata {}'.format(file) if (p.exists() and not skip_if_exists) else '')
        url = p.save()
        prt("Packaged saved to: {}".format(url))
        created = True
    elif p.exists():

        prt("Filesystem Package already exists")
        created = False
        url = join(p.package_path.path.rstrip('/'), DEFAULT_METATAB_FILE)

    return p, MetapackUrl(url, downloader=package_root.downloader), created


def make_csv_package(file, package_root, cache, env, skip_if_exists):
    assert package_root

    p = CsvPackageBuilder(file, package_root, callback=prt,  env=env)
    prt('Making CSV Package')
    if not p.exists() or not skip_if_exists:
        url = p.save()
        prt("Packaged saved to: {}".format(url))
        created = True
    elif p.exists():
        prt("CSV Package already exists")
        created = False
        url = p.package_path.path

    return p, MetapackUrl(url, downloader=package_root.downloader), created

def make_s3_package(file, package_root,  cache,  env,  skip_if_exists, acl='public-read'):

    assert package_root

    p = S3PackageBuilder(file, package_root, callback=prt,  env=env, acl=acl)

    if not p.exists() or not skip_if_exists:
        url = p.save()
        prt("Packaged saved to: {}".format(url))
        created = True
    elif p.exists():
        prt("S3 Filesystem Package already exists")
        created = False
        url = p.access_url

    return p, MetapackUrl(url, downloader=file.downloader), created



def update_name(mt_file, fail_on_missing=False, report_unchanged=True, force=False):

    if isinstance(mt_file, MetapackDoc):
        doc = mt_file
    else:
        doc = MetapackDoc(mt_file)

    o_name = doc.find_first_value("Root.Name", section=['Identity', 'Root'])

    updates = doc.update_name(force=force)

    for u in updates:
        prt(u)

    prt("Name is: ", doc.find_first_value("Root.Name", section=['Identity', 'Root']))

    if o_name != doc.find_first_value("Root.Name", section=['Identity', 'Root']) or force:
        write_doc(doc, mt_file)


def write_doc(doc, mt_file):
    """
    Write a Metatab doc to a CSV file, and update the Modified time
    :param doc:
    :param mt_file:
    :return:
    """


    import subprocess

    try:
        out = subprocess.run(['git', 'remote', 'show','origin'], stdout=subprocess.PIPE, timeout=6)\
            .stdout.decode('utf-8')

        fetchline = next(l.split() for l in out.splitlines() if 'Fetch' in l )
    except (TimeoutError, StopIteration, subprocess.TimeoutExpired):
        fetchline = None

    if fetchline:
        t = doc['Root'].get_or_new_term('GitUrl')
        t.value = fetchline[-1]

    u = parse_app_url(mt_file)

    if u.scheme == 'file':
        doc.write_csv(mt_file)
        return True
    else:
        return False
        #warn("Not writing back to url ", mt_file)


def process_schemas(mt_file, cache=None, clean=False):
    from rowgenerators import SourceError
    from requests.exceptions import ConnectionError

    if isinstance(mt_file, MetapackDoc):
        doc = mt_file
        write_doc_to_file = False
    else:
        doc = MetapackDoc(mt_file)
        write_doc_to_file = True

    try:
        if clean:
            doc['Schema'].clean()
        else:
            doc['Schema']

    except KeyError:
        doc.new_section('Schema', ['DataType', 'Altname', 'Description'])

    for r in doc['Resources'].find('Root.Resource'):

        schema_term = r.schema_term

        if schema_term:
            prt("Found table for '{}'; skipping".format(r.schema_name))
            continue

        path, name = extract_path_name(r.url)

        prt("Processing {}".format(r.url))

        slice = islice(r.row_generator, 500)

        si = SelectiveRowGenerator(slice,
                                   headers=[int(i) for i in r.get_value('headerlines', '0').split(',')],
                                   start=int(r.get_value('startline', 1)))

        try:
            ti = TypeIntuiter().run(si)
        except SourceError as e:
            warn("Failed to process '{}'; {}".format(path, e))
            continue
        except ConnectionError as e:
            warn("Failed to download '{}'; {}".format(path, e))
            continue

        table = doc['Schema'].new_term('Table', r.schema_name)

        prt("Adding table '{}' ".format(r.schema_name))

        for i, c in enumerate(ti.to_rows()):

            raw_alt_name = alt_col_name(c['header'], i)
            alt_name = raw_alt_name if raw_alt_name != c['header'] else ''

            t = table.new_child('Column', c['header'],
                            datatype=type_map.get(c['resolved_type'], c['resolved_type']),
                            altname=alt_name)

    if write_doc_to_file:
        write_doc(doc, mt_file)


def extract_path_name(ref):
    from os.path import splitext, basename, abspath

    uparts = parse_app_url(ref)

    ss = parse_app_url(ref)

    if not uparts.scheme:
        path = abspath(ref)
        name = basename(splitext(path)[0])
    else:
        path = ref

        v = ss.target_file if ss.target_file else uparts.path

        name = basename(splitext(v)[0])

    return path, name


def alt_col_name(name, i):
    import re

    if not name:
        return 'col{}'.format(i)

    return re.sub('_+', '_', re.sub('[^\w_]', '_', str(name)).lower()).rstrip('_')


type_map = {
    float.__name__: 'number',
    int.__name__: 'integer',
    six.text_type.__name__: 'string',
    six.binary_type.__name__: 'text',

}


class MetapackCliMemo(object):
    def __init__(self, args, downloader):
        from os import getcwd

        self.cwd = getcwd()
        self.args = args

        self.downloader = downloader

        self.cache = self.downloader.cache

        self.config = get_config()

        frag = ''

        if isinstance(args.metatabfile, (list, tuple)):
            args.metatabfile = args.metatabfile.pop(0)

        # Just the fragment was provided
        if args.metatabfile and args.metatabfile.startswith('#'):
            frag = args.metatabfile
            mtf = None
        else:
            frag = ''
            mtf = args.metatabfile

        # If could not get it from the args, Set it to the default file name in the current dir
        if not mtf:
            mtf = join(self.cwd, DEFAULT_METATAB_FILE)


        self.init_stage2(mtf, frag)


    def init_stage2(self, mtf, frag):

        self.frag = frag
        self.mtfile_arg = mtf + frag

        self.mtfile_url = MetapackUrl(self.mtfile_arg, downloader=self.downloader)

        # Find the target for a search URL
        if self.mtfile_url.scheme =='search':
            self.mtfile_url = parse_app_url(self.mtfile_arg, downloader=self.downloader).get_resource()

        self.resource = self.mtfile_url.resource_name

        self.package_url = self.mtfile_url.package_url
        self.mt_file = self.mtfile_url.metadata_url

        assert self.package_url.scheme == 'file'
        self.package_root = self.package_url.join(PACKAGE_PREFIX)
        assert self.package_root._downloader

    @property
    def doc(self):
        return MetapackDoc(self.mt_file)

def get_config():
    """Return a configuration dict"""
    from os import environ
    from os.path import expanduser
    from pathlib import Path
    import yaml

    def pexp(p):
        try:
            return Path(p).expanduser()
        except AttributeError:
            # python 3.4
            return Path(expanduser(p))

    paths = [ environ.get("METAPACK_CONFIG"),'~/.metapack.yaml','/etc/metapack.yaml']

    for p in paths:
        try:
            if pexp(p).exists():
                with  pexp(p).open() as f:
                    config = yaml.safe_load(f)
                    config['_loaded_from'] = str(pexp(p) )
                    return config
        except TypeError:
            pass


    return {}






from itertools import islice

from metapack.appurl import MetapackPackageUrl
from metapack.doc import EMPTY_SOURCE_HEADER
from metapack.exc import MetapackError, ResourceError
from metatab import Term
from rowgenerators import DownloadError, parse_app_url
from rowgenerators.rowpipe import RowProcessor


class Resource(Term):
    # These property names should return null if they aren't actually set.
    _common_properties = 'url name description schema'.split()

    def __init__(self, term, value, term_args=False, row=None, col=None, file_name=None, file_type=None,
                 parent=None, doc=None, section=None,
                 ):

        self.errors = {}  # Typecasting errors

        super().__init__(term, value, term_args, row, col, file_name, file_type, parent, doc, section)

    @property
    def base_url(self):
        """Base URL for resolving resource URLs"""

        if self.doc.package_url:
            return self.doc.package_url

        return self.doc._ref

    @property
    def env(self):
        """The execution context for rowprocessors and row-generating notebooks and functions. """
        from copy import copy

        env = copy(self.doc.env)

        env.update({
            # These become their own env vars when calling a program.
            'CACHE_DIR': self._doc._cache.getsyspath('/'),
            'RESOURCE_NAME': self.name,
            'RESOLVED_URL': str(self.resolved_url),
            'WORKING_DIR': str(self._doc.doc_dir),
            'METATAB_DOC': str(self._doc.ref),
            'METATAB_WORKING_DIR': str(self._doc.doc_dir),
            'METATAB_PACKAGE': str(self._doc.package_url)
        })

        env.update(self.all_props)

        return env

    @property
    def code_path(self):
        from .util import slugify
        from fs.errors import DirectoryExists

        sub_dir = 'resource-code/{}'.format(slugify(self.doc.name))
        try:
            self.doc.cache.makedirs(sub_dir)
        except DirectoryExists:
            pass

        return self.doc.cache.opendir(sub_dir).getsyspath(slugify(self.name) + '.py')

    @property
    def resolved_url(self):
        """Return a URL that properly combines the base_url and a possibly relative
        resource url"""

        if (self.term_is('root.sql') and not self.query) or (not self.term_is('root.sql') and  not self.url):
            return None

        # SQL Urls can be split into a Dsn part ( connection info ) and the SQL
        #  so the Sql urls have special handing for the references.
        dsns = {t.name:t.value for t in self.doc.find('Root.Dsn') }

        if self.term_is('root.sql'):
            u = parse_app_url(dsns[self.get_value('dsn')])
            u.sql = self.query

        else:
            u = parse_app_url(self.url)

        if u.scheme != 'file':
            # Hopefully means the URL is http, https, ftp, etc.
            return u
        elif u.resource_format == 'ipynb':

            # This shouldn't be a special case, but ...
            t = self.doc.package_url.inner.join_dir(self.url)
            t = t.as_type(type(u))
            t.fragment = u.fragment

            return t

        elif u.proto == 'metapack':
            return u.resource.resolved_url.get_resource().get_target()

        else:

            assert isinstance(self.doc.package_url, MetapackPackageUrl), (
                type(self.doc.package_url), self.doc.package_url)

            try:
                t = self.doc.package_url.resolve_url(self.url)

                # Also a hack
                t.scheme_extension = parse_app_url(self.url).scheme_extension

                # Another Hack!
                if not t.fragment and u.fragment:
                    t.fragment = u.fragment

                # Yet more hack!
                t = parse_app_url(str(t))

                return t

            except ResourceError as e:
                # This case happens when a filesystem packages has a non-standard metadata name
                # Total hack
                raise



    @property
    def inner(self):
        """For compatibility with the Appurl interface"""
        return self.resolved_url.get_resource().get_target().inner

    @property
    def parsed_url(self):
        return parse_app_url(self.url)

    def _name_for_col_term(self, c, i):

        altname = c.get_value('altname')
        name = c.value if c.value != EMPTY_SOURCE_HEADER else None
        default = "col{}".format(i)

        for n in [altname, name, default]:
            if n:
                return n

    @property
    def schema_name(self):
        """The value of the Name or Schema property"""
        return self.get_value('schema', self.get_value('name'))

    @property
    def schema_table(self):
        """Deprecated. Use schema_term()"""
        return self.schema_term

    @property
    def schema_term(self):
        """Return the Table term for this resource, which is referenced either by the `table` property or the
        `schema` property"""

        if not self.name:
            raise MetapackError("Resource for url '{}' doe not have name".format(self.url))

        t = self.doc.find_first('Root.Table', value=self.get_value('name'))
        frm = 'name'

        if not t:
            t = self.doc.find_first('Root.Table', value=self.get_value('schema'))
            frm = 'schema'

        if not t:
            frm = None

        return t

    @property
    def headers(self):
        """Return the headers for the resource. Returns the AltName, if specified; if not, then the
        Name, and if that is empty, a name based on the column position. These headers
        are specifically applicable to the output table, and may not apply to the resource source. FOr those headers,
        use source_headers"""

        t = self.schema_term

        if t:
            return [self._name_for_col_term(c, i)
                    for i, c in enumerate(t.children, 1) if c.term_is("Table.Column")]
        else:
            return None

    @property
    def source_headers(self):
        """"Returns the headers for the resource source. Specifically, does not include any header that is
        the EMPTY_SOURCE_HEADER value of _NONE_"""

        t = self.schema_term

        if t:
            return [self._name_for_col_term(c, i)
                    for i, c in enumerate(t.children, 1) if c.term_is("Table.Column")
                    and c.get_value('name') != EMPTY_SOURCE_HEADER
                    ]
        else:
            return None

    def columns(self):

        try:
            # For resources that are metapack packages.
            r =  self.resolved_url.resource.columns()
            yield from r
        except AttributeError:
            pass

        t = self.schema_term

        if not t:
            return

        for i, c in enumerate(t.children):

            if c.term_is("Table.Column"):
                p = c.all_props
                p['pos'] = i
                p['name'] = c.value
                p['header'] = self._name_for_col_term(c, i)

                yield p

    def row_processor_table(self, ignore_none=False):
        """Create a row processor from the schema, to convert the text values from the
        CSV into real types"""
        from rowgenerators.rowpipe import Table

        type_map = {
            None: None,
            'string': 'str',
            'text': 'str',
            'number': 'float',
            'integer': 'int'
        }

        def map_type(v):
            return type_map.get(v, v)

        if self.schema_term:

            t = Table(self.get_value('name'))

            col_n = 0

            for c in self.schema_term.children:

                if ignore_none and c.name == EMPTY_SOURCE_HEADER:
                    continue

                if c.term_is('Table.Column'):
                    t.add_column(self._name_for_col_term(c, col_n),
                                 datatype=map_type(c.get_value('datatype')),
                                 valuetype=map_type(c.get_value('valuetype')),
                                 transform=c.get_value('transform'),
                                 width=c.get_value('width')
                                 )
                    col_n += 1

            return t

        else:
            return None

    @property
    def row_generator(self):
        from rowgenerators import get_generator

        self.doc.set_sys_path()  # Set sys path to package 'lib' dir in case of python function generator

        ru = self.resolved_url

        try:
            resource = ru.resource # For Metapack urls

            return resource.row_generator
        except AttributeError:
            pass

        ut = ru.get_resource().get_target()

        # Encoding is supposed to be preserved in the URL but isn't
        source_url = parse_app_url(self.url)

        # soruce_url will be None for Sql terms.
        ut.encoding = source_url.encoding if source_url else self.get_value('encoding')

        table = self.row_processor_table()

        g = get_generator(ut, table=table, resource=self,
                          doc=self._doc, working_dir=self._doc.doc_dir,
                          env=self.env)

        assert g, ut

        return g

    def _get_header(self):
        """Get the header from the deinfed header rows, for use  on references or resources where the schema
        has not been run"""

        try:
            header_lines = [int(e) for e in str(self.get_value('headerlines', 0)).split(',')]
        except ValueError as e:
            header_lines = [0]

        # We're processing the raw datafile, with no schema.
        header_rows = islice(self.row_generator, min(header_lines), max(header_lines) + 1)

        from tableintuit import RowIntuiter
        headers = RowIntuiter.coalesce_headers(header_rows)

        return headers

    def __iter__(self):
        """Iterate over the resource's rows"""

        headers = self.headers

        # There are several args for SelectiveRowGenerator, but only
        # start is really important.
        try:
            start = int(self.get_value('startline', 1))
        except ValueError as e:
            start = 1

        if headers:  # There are headers, so use them, and create a RowProcess to set data types
            yield headers

            base_row_gen = self.row_generator

            assert base_row_gen is not None

            assert type(self.env) == dict

            rg = RowProcessor(islice(base_row_gen, start, None),
                              self.row_processor_table(),
                              source_headers=self.source_headers,
                              manager = self,
                              env=self.env,
                              code_path=self.code_path)

        else:
            headers = self._get_header()  # Try to get the headers from defined header lines

            yield headers
            rg = islice(self.row_generator, start, None)

        yield from rg

        try:
            self.errors = rg.errors if rg.errors else {}
        except AttributeError:
            self.errors = {}

    @property
    def iterdict(self):
        """Iterate over the resource in dict records"""
        from collections import OrderedDict

        headers = None

        for row in self:

            if headers is None:
                headers = row
                continue

            yield OrderedDict(zip(headers, row))

    @property
    def iterrows(self):
        """Iterate over the resource as row proxy objects, which allow acessing colums as attributes"""

        from rowgenerators.rowproxy import RowProxy

        row_proxy = None

        headers = None

        for row in self:

            if not headers:
                headers = row
                row_proxy = RowProxy(headers)
                continue

            yield row_proxy.set_row(row)

    @property
    def json_headers(self):
        return [(c['pos'], c.get('json') or c['header']) for c in self.columns() if c.get('json')]

    @property
    def iterstruct(self):
        """Yield data structures built from the JSON header specifications in a table"""
        from rowgenerators.rowpipe.json import add_to_struct

        json_headers = self.json_header

        for row in islice(self, 1, None): # islice skips header
            d = {}
            for pos, jh in json_headers:
                add_to_struct(d, jh, row[pos])
            yield d

    def iterjson(self, *args, **kwargs):
        """Yields the data structures from iterstruct as JSON strings"""
        from rowgenerators.rowpipe.json import VTEncoder
        import json

        if 'cls' not in kwargs:
            kwargs['cls'] = VTEncoder

        for s in self.iterstruct:
            yield (json.dumps(s, *args, **kwargs))

    def iteryaml(self, *args, **kwargs):
        """Yields the data structures from iterstruct as YAML strings"""
        from rowgenerators.rowpipe.json import VTEncoder
        import yaml

        if 'cls' not in kwargs:
            kwargs['cls'] = VTEncoder

        for s in self.iterstruct:
            yield (yaml.safe_dump(s))

    def dataframe(self, limit=None):
        """Return a pandas datafrome from the resource"""

        from metapack.jupyter.pandas import MetatabDataFrame

        rg = self.row_generator

        # Maybe generator has it's own Dataframe method()
        try:
            return rg.dataframe()
        except AttributeError:
            pass

        # Just normal data, so use the iterator in this object.
        headers = next(islice(self, 0, 1))
        data = islice(self, 1, None)

        df = MetatabDataFrame(list(data), columns=headers, metatab_resource=self)

        self.errors = df.metatab_errors = rg.errors if hasattr(rg, 'errors') and rg.errors else {}

        return df

    def geoframe(self):
        """Return a Geo dataframe"""

        return self.dataframe().geo

    def read_csv(self, *args, **kwargs):
        """Fetch the target and pass through to pandas.read_csv

        Don't provide the first argument of read_csv(); it is supplied internally.
        """
        import pandas

        t = self.resolved_url.get_resource().get_target()

        return pandas.read_csv(t.path, *args, **kwargs)

    def read_fwf(self, *args, **kwargs):
        """Fetch the target and pass through to pandas.read_fwf.

        Don't provide the first argument of read_fwf(); it is supplied internally. """
        import pandas

        t = self.resolved_url.get_resource().get_target()

        return pandas.read_fwf(t.path, *args, **kwargs)

    def readlines(self):
        """Load the target, open it, and return the result from readlines()"""

        t = self.resolved_url.get_resource().get_target()
        with open(t.path) as f:
            return f.readlines()

    def petl(self, *args, **kwargs):
        """Return a PETL source object"""
        import petl

        t = self.resolved_url.get_resource().get_target()

        if t.target_format == 'txt':
            return petl.fromtext(t.path, *args, **kwargs)
        elif t.target_format == 'csv':
            return petl.fromcsv(t.path, *args, **kwargs)
        else:
            raise Exception("Can't handle")

    def _repr_html_(self):


        try:
            return self.sub_resource._repr_html_()
        except AttributeError:
            pass
        except DownloadError:
            pass

        return (
                   "<h3><a name=\"resource-{name}\"></a>{name}</h3><p><a target=\"_blank\" href=\"{url}\">{url}</a></p>" \
                       .format(name=self.name, url=self.resolved_url)) + \
               "<table>\n" + \
               "<tr><th>Header</th><th>Type</th><th>Description</th></tr>" + \
               '\n'.join(
                   "<tr><td>{}</td><td>{}</td><td>{}</td></tr> ".format(c.get('header', ''),
                                                                        c.get('datatype', ''),
                                                                        c.get('description', ''))
                   for c in self.columns()) + \
               '</table>'

    @property
    def markdown(self):

        from .html import ckan_resource_markdown
        return ckan_resource_markdown(self)


class Reference(Resource):
    @property
    def env(self):
        e = super().env
        e['reference'] = self
        return e

    def __iter__(self):
        """Iterate over the resource's rows"""

        try:
            # For Metapack references
            yield from self.resolved_url.resource
        except AttributeError:
            yield from self.row_generator

class Distribution(Term):
    @property
    def type(self):

        # The following order is really important.
        if self.package_url.target_format == 'xlsx':
            return 'xlsx'
        elif self.package_url.resource_format == 'zip':
            return "zip"
        elif self.metadata_url.target_file == 'metadata.csv':
            return 'fs'
        elif self.package_url.target_format == 'csv':
            return "csv"

        else:
            return "unk"

    @property
    def package_url(self):
        from metapack import MetapackPackageUrl
        return MetapackPackageUrl(self.value, downloader=self.doc.downloader)

    @property
    def metadata_url(self):
        from metapack import MetapackDocumentUrl
        return MetapackDocumentUrl(self.value, downloader=self.doc.downloader)



# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""Support for PANDAS dataframes"""

from pandas import DataFrame, Series



class MetatabSeries(Series):

    _metadata = [ 'metatab_resource', 'name', '_description', '_schema'] # Name is defined in the parent

    @property
    def _constructor(self):
        return MetatabSeries

    @property
    def _constructor_expanddim(self):
        return MetatabDataFrame

    @property
    def column(self):
        """Return the information about the column"""
        raise NotImplementedError()

    @property
    def description(self): # b/c the _metadata elements aren't created until assigned
        try:
            return self._description
        except AttributeError:
            return None

    @description.setter
    def description(self, v):
        self._description = v


class MetatabDataFrame(DataFrame):

    _metadata = [ 'metatab_resource','metatab_errors', '_name',  '_title']

    def __init__(self, data=None, index=None, columns=None, dtype=None, copy=False, metatab_resource=None):

        self.metatab_resource = metatab_resource
        self.metatab_errors = {}

        super(MetatabDataFrame, self).__init__(data, index, columns, dtype, copy)

    @property
    def geo(self):
        """Return a geopandas dataframe"""
        import geopandas as gpd
        from shapely.geometry.polygon import BaseGeometry
        from shapely.wkt import loads

        gdf = gpd.GeoDataFrame(self)

        first = next(gdf.iterrows())[1].geometry

        if isinstance(first, str):
            shapes = [ loads(row['geometry']) for i, row in gdf.iterrows()]

        elif not isinstance(first, BaseGeometry):
            # If we are reading a metatab package, the geometry column's type should be
            # 'geometry' which will give the geometry values class type of
            # rowpipe.valuetype.geo.ShapeValue. However, there are other
            # types of objects that have a 'shape' property.

            shapes = [row['geometry'].shape for i, row in gdf.iterrows()]

        else:
            shapes = gdf['geometry']

        gdf['geometry'] = gpd.GeoSeries(shapes)
        gdf.set_geometry('geometry')
        return gdf

    @property
    def _constructor(self):

        return MetatabDataFrame

    @property
    def _constructor_sliced(self):
        return MetatabSeries

    def _getitem_column(self, key):
        c = super(MetatabDataFrame, self)._getitem_column(key)
        c.metatab_resource = self.metatab_resource
        return c

    def copy(self, deep=True):
        df = super().copy(deep)

        for c in df.columns:
            df[c].__class__ = MetatabSeries



        return df

    @property
    def title(self):  # b/c the _metadata elements aren't created until assigned
        try:
            return self._title
        except AttributeError:
            return ''

    @title.setter
    def title(self, v):
        self._title = v

    @property
    def name(self):  # b/c the _metadata elements aren't created until assigned
        try:
            return self._name
        except AttributeError:
            return ''

    @name.setter
    def name(self, v):
        self._name = v

    @property
    def rows(self):
        """Yield rows like a partition does, with a header first, then rows. """

        yield [self.index.name] + list(self.columns)

        for t in self.itertuples():
            yield list(t)





from os import walk
from os.path import join
import zipfile
from metapack.util import slugify

from .core import PackageBuilder


class ZipPackageBuilder(PackageBuilder):
    """A Zip File package"""

    type_code = 'zip'

    def __init__(self, source_ref=None, package_root=None,  callback=None, env=None):

        super().__init__(source_ref, package_root, callback, env)

        self.package_path, self.cache_path = self.make_package_path(self.package_root, self.package_name)

    @classmethod
    def make_package_path(cls, package_root, package_name):

        cache_path = package_name + ".zip"

        package_path = package_root.join(cache_path)

        return package_path, cache_path

    def save(self, path=None):

        self.check_is_ready()

        root_dir = slugify(self.doc.find_first_value('Root.Name'))

        self.prt("Creating ZIP Package at '{}' from filesystem package at '{}'"
                 .format(self.package_path, self.source_dir))

        self.zf = zipfile.ZipFile(self.package_path.path, 'w', zipfile.ZIP_DEFLATED)


        for root, dirs, files in walk(self.source_dir):
            for f in files:
                source = join(root, f)
                rel = source.replace(self.source_dir,'').strip('/')
                dest = join(root_dir, rel)

                self.zf.write(source,dest)

        self.zf.close()

        return self.package_path


# -*- coding: utf-8 -*
# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""

jupyter nbextension install --py metatab.jupyter.magic

"""

from __future__ import print_function

import logging
import os
import shlex
import sys
from collections import OrderedDict
from os import makedirs, getcwd
from os.path import join, abspath, dirname, exists, normpath
from warnings import warn

import docopt
from IPython import get_ipython
from IPython.core.magic import Magics, magics_class, line_magic, cell_magic, line_cell_magic
from IPython.core.magic_arguments import (argument, magic_arguments,
                                          parse_argstring)
from IPython.display import  display, HTML, Latex

from rowgenerators import Downloader, parse_app_url
from metapack import MetapackDoc
from metapack.appurl import MetapackPackageUrl
from metapack.cli.core import process_schemas
from metapack.html import bibliography, data_sources
from metatab import TermParser
from metatab.generate import TextRowGenerator

logger = logging.getLogger('user')
logger_err = logging.getLogger('cli-errors')
doc_logger = logging.getLogger('doc')
debug_logger = logging.getLogger('debug_logger')

MT_DOC_VAR = 'mt_pkg' # Namespace name for the metatab document.

@magics_class
class MetatabMagic(Magics):

    """Magics for using Metatab in Jupyter Notebooks

    """



    @property
    def mt_doc(self):
        """Return the current metatab document, which must be created with either %%metatab
        or %mt_load_package"""



        if MT_DOC_VAR not in self.shell.user_ns:

            package_url = MetapackPackageUrl("metapack+file:" + os.getcwd() + '/', downloader=Downloader())

            self.shell.user_ns[MT_DOC_VAR] = \
                MetapackDoc(TextRowGenerator("Declare: metatab-latest\n"), package_url=package_url)

            inline_doc = self.shell.user_ns[MT_DOC_VAR]

            if not 'Resources' in inline_doc:
                inline_doc.new_section('Resources', ['Name', 'Description'])
            if not 'Resources' in inline_doc:
                inline_doc.new_section('References', ['Name', 'Description'])

            # Give all of the sections their standard args, to make the CSV versions of the doc
            # prettier
            for name, s in inline_doc.sections.items():
                try:
                    s.args = inline_doc.decl_sections[name.lower()]['args']
                except KeyError:
                    pass

        return self.shell.user_ns[MT_DOC_VAR]

    @property
    def package_dir(self):
        """Return the current metatab document, which must be created with either %%metatab
        or %mt_load_package"""

        return self.shell.user_ns['_package_dir']

    def add_term_lines(self, text):


        assert 'root.reference' in TermParser.term_classes

        tp = TermParser(TextRowGenerator(text), resolver= self.mt_doc.resolver, doc= self.mt_doc)

        self.mt_doc.load_terms(tp)


    def clean_doc(self, doc):

        # Some sections have terms that are unique on Name

        for sec in ['Resources','References','Bibliography', 'Schema', 'Contacts']:
            try:
                seen = set()
                for t in doc[sec]:
                    if t.name in seen:
                        doc.remove_term(t)
                    seen.add(t.name)
            except KeyError:
                pass

        # For the Root, just make every term unique name term name, keeping the last one

        seen = {}
        for t in list(doc['Root']):
            if t.term in seen:
                extant = seen[t.term]
                doc.remove_term(extant)
            seen[t.term] = t

    @magic_arguments()
    @argument('-s', '--show', help='After loading, display the document', action='store_true')
    @argument('-p', '--package_dir', help='Set the directory where the package will be created')
    @cell_magic
    def metatab(self, line, cell):
        """Process a cell of Metatab data, in line format. Stores document in the `mt_pkg` variable
        """

        args = parse_argstring(self.metatab, line)

        inline_doc = self.mt_doc

        self.add_term_lines(cell)

        extant_identifier = inline_doc.get_value('Root.Identifier')
        extant_name = inline_doc.get_value('Root.Name')

        inline_doc.update_name(force=True, create_term=True)

        process_schemas(inline_doc)

        self.clean_doc(inline_doc)

        if args.show:
            for l in inline_doc.lines:
                print(': '.join(str(e) for e in l))

        if args.package_dir:
            self.shell.user_ns['_package_dir'] = abspath(join(getcwd(), args.package_dir))
        else:
            self.shell.user_ns['_package_dir'] = join(getcwd(), inline_doc.get_value('Root.Name'))

        if extant_identifier != inline_doc.get_value('Root.Identifier'):
            print("Identifier updated.  Set 'Identifier: {}'  in document".format(inline_doc.get_value('Root.Identifier')))

        if extant_name != inline_doc.get_value('Root.Name'):
            print("Name Changed. Set 'Name: {}'  in document".format(inline_doc.get_value('Root.Name')))


    @magic_arguments()
    @argument('-s', '--source', help='Force opening the source package', action='store_true')
    @line_magic
    def mt_open_package(self, line):
        """Find the metatab file for this package, open it, and load it into the namespace. """

        from metapack.jupyter.ipython import open_package

        args = parse_argstring(self.mt_open_package, line)
        self.shell.user_ns[MT_DOC_VAR] = open_package(self.shell.user_ns)


        if self.mt_doc.package_url:
            u = parse_app_url(self.mt_doc.package_url)

    @line_magic
    def mt_import_terms(self, line):
        """Import the value of some Metatab terms into the notebook namespace """

        mt_terms = {}

        doc = self.mt_doc

        mt_terms['root.title'] = doc['Root'].find_first_value('Root.Title')
        mt_terms['root.description'] = doc['Root'].find_first_value('Root.Description')
        mt_terms['root.name'] = doc['Root'].find_first_value('Root.Name')
        mt_terms['root.identifier'] = doc['Root'].find_first_value('Root.Identifier')

        mt_terms['contacts'] = []

        for t in doc.get_section('Contacts', []):
            d = t.as_dict()
            d['type'] = t.record_term_lc
            mt_terms['contacts'].append(d)

        mt_terms['bibliography'] = []

        for t in doc.get_section('Bibliography', []):
            d = t.as_dict()
            d['type'] = t.record_term_lc
            mt_terms['bibliography'].append(d)

        mt_terms['references'] = []

        for t in doc.get_section('References', []):
            d = t.as_dict()
            d['type'] = t.record_term_lc
            mt_terms['references'].append(d)

        print('mt_terms')

        self.shell.user_ns['mt_terms'] = mt_terms

    @line_magic
    def mt_process_schemas(self, line):
        """Add Schema entries for resources to the metatab file. Does not write the doc file"""

        process_schemas(self.mt_doc)

    @magic_arguments()
    @argument('package_dir', help='Package directory')
    @argument('--feather', help='Use feather for serialization', action='store_true' )
    @line_magic
    def mt_materialize_all(self, line):
        """Materialize all of the dataframes that has been previoulsly added with mt_add_dataframe
        """

        from json import dumps

        from metapack.rowgenerator import PandasDataframeSource
        import csv

        materialized = []

        args = parse_argstring(self.mt_materialize_all, line)

        try:
            cache = self.mt_doc._cache
        except KeyError:
            cache = Downloader().cache

        for df_name, ref in self.shell.user_ns.get('_material_dataframes',{}).items():

            u = parse_app_url(args.package_dir).join(ref)

            path = u.path

            if not exists(dirname(path)):
                makedirs(dirname(path))

            df = self.shell.user_ns[df_name].fillna('')

            if args.feather:
                path = path.replace('.csv','.feather')
                df.to_feather(path)
            else:
                gen = PandasDataframeSource(u, df, cache=cache)
                with open(path, 'w') as f:
                    w = csv.writer(f)
                    w.writerows(gen)

            materialized.append({
                'df_name': df_name,
                'path': path,
                'ref':ref
            })

        # Show information about the dataframes that were materialized, so it can be harvested later
        print(dumps(materialized, indent=4))

    @magic_arguments()
    @argument('df_name', help='Dataframe name')
    @argument('dir', help='Path to output directory. Created if it does not exist')
    @argument('--feather', help='Materialize with feather', action="store_true")
    @line_magic
    def mt_materialize(self, line):
        """Materialize all of the dataframes that has been previoulsly added with mt_add_dataframe
        """

        from json import dumps
        from rowgenerators import get_cache
        from metapack.rowgenerator import PandasDataframeSource
        import csv
        from os.path import join

        args = parse_argstring(self.mt_materialize, line)

        try:
            cache = self.mt_doc._cache
        except KeyError:
            cache = get_cache()

        if not exists(args.dir):
            makedirs(args.dir)

        if args.feather:
            path = join(args.dir, args.df_name+".feather")
            df = self.shell.user_ns[args.df_name]
            df.to_feather(path)
        else:
            path = join(args.dir, args.df_name + ".csv")
            df = self.shell.user_ns[args.df_name].fillna('')
            gen = PandasDataframeSource(parse_app_url(path), df, cache=cache)

            with open(path, 'w') as f:
                w = csv.writer(f)
                w.writerows(gen)

        print(dumps({
            'df_name': args.df_name,
            'path': path,

        }, indent=4))


    @line_magic
    def mt_show_metatab(self, line):
        """Dump the metatab file to the output, so it can be harvested after the notebook is executed"""

        try:
            for line in self.mt_doc.lines:

                if  line[1]: # Don't display "None"
                    print(': '.join(line))
        except KeyError:
            pass

    @magic_arguments()
    @argument('-m','--materialize', help='Save the data for the dataframe during package conversion', action="store_true")
    @argument('-n','--name', help='Metadata reference name of the dataframe')
    @argument('-t','--title', help='Title of the dataframe')
    @argument('-d', '--dump', help='Dump example schema for the dataframe. Implied when used as line magic', action="store_true")
    @argument('--feather', help='Materialize with feather', action="store_true")
    @argument('dataframe_name', nargs=1, help='Variable name of the dataframe name')
    @line_cell_magic
    def mt_add_dataframe(self, line, cell=''):
        """Add a dataframe to a metatab document's data files

        """
        from metapack.jupyter.core import process_schema
        from metatab.exc import ParserError

        args = parse_argstring(self.mt_add_dataframe, line)

        if not cell:
            is_line = True
            args.dump = True
        else:
            is_line = False

        dataframe_name = args.dataframe_name[0]

        if not '_material_dataframes' in self.shell.user_ns:
            self.shell.user_ns['_material_dataframes'] = {}

        df = self.shell.user_ns[dataframe_name]

        try:
            cell_doc = MetapackDoc(TextRowGenerator("Declare: metatab-latest\n" + cell))
        except ParserError as e:
            warn('Failed to parse Metatab in cell: {} '.format(e))
            return

        cell_table = cell_doc.find_first('Root.Table')

        if cell_table and ( args.name or args.title ):
            warn("The name and title arguments are ignored when the cell includes a Metatab table definition")

        if cell_table:
            name = cell_table.get_value('name')
            title = cell_table.get_value('title','')
            description = cell_table.get_value('description','')
        else:
            name = None
            title = ''
            description = ''


        if not name:
            name = args.name or dataframe_name;

        if not title:
            title = args.title or dataframe_name

        if not name:
            warn("Name must be set with .name property, or --name option".format(args['<dataframe_name>']))
            return

        title = title.strip("'").strip('"')

        try:
            doc = self.mt_doc
        except KeyError:
            doc = None

        if args.materialize:
            ref = 'file:data/{}.csv'.format(name)
            self.shell.user_ns['_material_dataframes'][dataframe_name] = ref

        elif doc is not None:
            ref = 'ipynb:notebooks/{}.ipynb#{}'.format( doc.as_version(None), dataframe_name)

        else:
            ref = None

        table = None

        resource_term = None

        ##
        ## First, process the schema, extracting the columns from the dataframe.
        ##

        if doc and ref:
            if not 'Resources' in doc:
                doc.new_section('Resources')

            resource_term = doc['Resources'].get_or_new_term("Root.Datafile", ref)

            resource_term['name'] = name
            resource_term['title'] = title
            resource_term['description'] = description

            df = df.reset_index()

            table = process_schema(doc, doc.resource(name), df)

            if not table:
                table = doc['Schema'].find_first('Root.Table', name)

        ##
        ## Next, apply the names from  table description from the cell
        ##

        if cell_table:

            cols_by_name = {c.name:c for c in cell_table.find('Table.Column') }

            for i,c in enumerate(table.find('Table.Column')):

                cell_column = cols_by_name.get(c.name)
                try:
                    cell_col_by_pos = list(cols_by_name.values())[i]
                except KeyError:
                    cell_col_by_pos = None
                except IndexError:
                    cell_col_by_pos = None

                if cell_column:
                    c.description = cell_column.description
                    c.name = cell_column.name
                elif cell_col_by_pos:
                    c.description = cell_col_by_pos.description
                    c.name = cell_col_by_pos.name


        if args.dump and table:
            print("Table:", resource_term.name)

            if resource_term and resource_term.title:
                print("Table.Title:", resource_term.get_value('title'))
                print("Table.Description:", resource_term.get_value('description') if resource_term.get_value('description')  else '')

            for c in table.find('Table.Column'):
                print("Table.Column:", c.name)
                print("  .Datatype:", c.datatype)
                print("  .Description:",  c.description or '')

            if is_line:
                print("\nCopy the above into the cell, and change to a cell magic, with '%%' ")


    @magic_arguments()
    @argument('--format', help="Format, html or latex. Defaults to 'all' ", default='all', nargs='?', )
    @argument('converters', help="Class names for citation converters", nargs='*', )
    @line_magic
    def mt_bibliography(self, line):
        """Display, as HTML, the bibliography for the metatab document. With no argument,
         concatenate all doc, or with an arg, for only one. """

        args = parse_argstring(self.mt_bibliography, line)

        def import_converter(name):
            components = name.split('.')
            mod = __import__(components[0])
            for comp in components[1:]:
                mod = getattr(mod, comp)
            return mod

        converters = [ import_converter(e) for e in args.converters ]

        if args.format == 'html' or args.format == 'all':
            display(HTML(bibliography(self.mt_doc, converters=converters, format='html')))

        if args.format == 'latex' or args.format == 'all':
            display(Latex(bibliography(self.mt_doc, converters=converters, format='latex')))

    @magic_arguments()
    @argument('--format', help="Format, html or latex. Defaults to 'all' ", default='all', nargs='?', )
    @argument('converters', help="Class names for citation converters", nargs='*', )
    @line_magic
    def mt_data_references(self, line):
        """Display, as HTML, the bibliography for the metatab document. With no argument,
         concatenate all doc, or with an arg, for only one. """

        args = parse_argstring(self.mt_bibliography, line)

        def import_converter(name):
            components = name.split('.')
            mod = __import__(components[0])
            for comp in components[1:]:
                mod = getattr(mod, comp)
            return mod

        converters = [import_converter(e) for e in args.converters]

        if args.format == 'html' or args.format == 'all':
            display(HTML(data_sources(self.mt_doc, converters=converters, format='html')))

        if args.format == 'latex' or args.format == 'all':
            display(Latex(data_sources(self.mt_doc, converters=converters, format='latex')))

    @magic_arguments()
    @argument('lib_dir', help='Directory', nargs='?',)
    @line_magic
    def mt_lib_dir(self, line):
        """Declare a source code directory and add it to the sys path

        The argument may be a directory, a URL to a Metatab ZIP archive, or a reference to a
        Root.Reference or Root.Resource term that references a Metatab ZIP Archive

        If lib_dir is not specified, it defaults to 'lib'

        If lib_dir is a directory, the target is either at the same level as the CWD or
        one level up.

        If lib_dir is a URL, it must point to a Metatab ZIP archive that has an interal Python
        package directory. The URL may hav path elements after the ZIP archive to point
        into the ZIP archive. For instance:

            %mt_lib_dir http://s3.amazonaws.com/library.metatab.org/ipums.org-income_homevalue-5.zip

        If lib_dir is anything else, it is a reference to the name of a Root.Reference or Root.Resource term that
        references a Metatab ZIP Archive. For instance:

            %%metatab
            ...
            Section: References
            Reference: metatab+http://s3.amazonaws.com/library.metatab.org/ipums.org-income_homevalue-5.zip#income_homeval
            ...


            %mt_lib_dir incv
            from lib.incomedist import *

        """

        from os.path import splitext, basename

        args = parse_argstring(self.mt_lib_dir, line) # Its a normal string

        if not args.lib_dir:
            lib_dir = 'lib'

        else:
            lib_dir = args.lib_dir

        if not '_lib_dirs' in self.shell.user_ns:
            self.shell.user_ns['_lib_dirs'] = set()

        u = parse_app_url(lib_dir)

        # Assume files are actually directories
        # This clause will pickup both directories and reference names, but ref names should not
        # exist as directories.
        if u.proto == 'file':

            lib_dir = normpath(lib_dir).lstrip('./')

            for path in [ abspath(lib_dir), abspath(join('..',lib_dir))]:
                if exists(path) and path not in sys.path:
                    sys.path.insert(0,path)
                    self.shell.user_ns['_lib_dirs'].add(lib_dir)
                    return

        # Assume URLS are to Metapack packages on the net
        if (u.scheme == 'https' or u.scheme == 'http'):

            # The path has to be a Metatab ZIP archive, and the root directory must be the same as
            # the name of the path

            r = u.get_resource()

            pkg_name, _ = splitext(basename(r.path))

            lib_path = r.join(pkg_name).path

            if lib_path not in sys.path:
                sys.path.insert(0,lib_path)

        # Assume anything else is a Metatab Reference term name
        elif self.mt_doc.find_first('Root.Reference', name=lib_dir) or self.mt_doc.resource(lib_dir) :

            r = self.mt_doc.find_first('Root.Reference', name=lib_dir) or self.mt_doc.resource(lib_dir)

            ur = parse_app_url(r.url).inner

            return self.mt_lib_dir(str(ur))

        else:
            logger.error("Can't find library directory: '{}' ".format(lib_dir))


    @line_magic
    def mt_show_libdirs(self, line):
        """Dump the list of lib dirs as JSON"""
        import json

        if '_lib_dirs' in self.shell.user_ns:
            print(json.dumps(list(self.shell.user_ns['_lib_dirs'])))
        else:
            print(json.dumps([]))


def load_ipython_extension(ipython):
    # In order to actually use these magics, you must register them with a
    # running IPython.  This code must be placed in a file that is loaded once
    # IPython is up and running:
    ip = get_ipython()
    # You can register the class itself without instantiating it.  IPython will
    # call the default constructor on it.
    ip.register_magics(MetatabMagic)

    # init_logging()


def unload_ipython_extension(ipython):
    # If you want your extension to be unloadable, put that logic here.
    pass



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
Support for running notebooks converted to scripts
"""

import errno
import signal
import sys
import time
from subprocess import Popen, PIPE

from IPython.core import prefilter, magic_arguments
from IPython.core.magic import Magics, magics_class, line_magic, cell_magic
from IPython.utils import py3compat
from IPython.utils.process import arg_split
from IPython.utils.text import DollarFormatter
from metatab.ipython import caller_locals
from os import system
from .magic import MetatabMagic
from IPython.core.magics.script import script_args

class ScriptIPython(object):
    """A Fake IPython object to supportin running some magics from converted scripts"""

    def __init__(self, user_ns):
        self.user_ns = user_ns
        self.mm = MetatabMagic(shell=self)
        pass

    def magic(self, arg_s):
        magic_name, _, magic_arg_s = arg_s.partition(' ')
        magic_name = magic_name.lstrip(prefilter.ESC_MAGIC)
        return self.run_line_magic(magic_name, magic_arg_s)

    def run_line_magic(self, magic_name, line):

        if hasattr(self.mm, magic_name):
            # Run any magics for metatab
            f = getattr(self.mm, magic_name)
            f(line)
            return

        elif magic_name == 'matplotlib':
            # Don't let matplotlib open a display window.
            import matplotlib
            matplotlib.use('AGG')


    def run_cell_magic(self, magic_name, line, cell):
        """Run a limited number of magics from scripts, without IPython"""

        if magic_name == 'bash':
            self.shebang("bash", cell)
        elif magic_name == 'metatab':
            self.mm.metatab(line, cell)


    # Stolen from IPython  distribution
    def system_piped(self, cmd):
        """Call the given cmd in a subprocess, piping stdout/err

        Parameters
        ----------
        cmd : str
          Command to execute (can not end in '&', as background processes are
          not supported.  Should not be a command that expects input
          other than simple text.
        """
        if cmd.rstrip().endswith('&'):
            # this is *far* from a rigorous test
            # We do not support backgrounding processes because we either use
            # pexpect or pipes to read from.  Users can always just call
            # os.system() or use ip.system=ip.system_raw
            # if they really want a background process.
            raise OSError("Background processes not supported.")

        # we explicitly do NOT return the subprocess status code, because
        # a non-None value would trigger :func:`sys.displayhook` calls.
        # Instead, we store the exit_code in user_ns.
        self.user_ns['_exit_code'] = system(self.var_expand(cmd, depth=1))

    # Stolen from IPython  distribution
    def var_expand(self, cmd, depth=0, formatter=DollarFormatter()):
        """Expand python variables in a string.

        The depth argument indicates how many frames above the caller should
        be walked to look for the local namespace where to expand variables.

        The global namespace for expansion is always the user's interactive
        namespace.
        """
        ns = self.user_ns.copy()
        try:
            frame = sys._getframe(depth + 1)
        except ValueError:
            # This is thrown if there aren't that many frames on the stack,
            # e.g. if a script called run_line_magic() directly.
            pass
        else:
            ns.update(frame.f_locals)

        try:
            # We have to use .vformat() here, because 'self' is a valid and common
            # name, and expanding **ns for .format() would make it collide with
            # the 'self' argument of the method.
            cmd = formatter.vformat(cmd, args=[], kwargs=ns)
        except Exception:
            # if formatter couldn't format, just let it go untransformed
            pass
        return cmd

    system = system_piped

    @magic_arguments.magic_arguments()
    @script_args
    @cell_magic("script")
    def shebang(self, line, cell):
        """Run a cell via a shell command

        The `%%script` line is like the #! line of script,
        specifying a program (bash, perl, ruby, etc.) with which to run.

        The rest of the cell is run by that program.

        Examples
        --------
        ::

            In [1]: %%script bash
               ...: for i in 1 2 3; do
               ...:   echo $i
               ...: done
            1
            2
            3
        """
        argv = arg_split(line, posix=not sys.platform.startswith('win'))
        args, cmd = self.shebang.parser.parse_known_args(argv)

        try:
            p = Popen(cmd, stdout=PIPE, stderr=PIPE, stdin=PIPE)
        except OSError as e:
            if e.errno == errno.ENOENT:
                print("Couldn't find program: %r" % cmd[0])
                return
            else:
                raise

        if not cell.endswith('\n'):
            cell += '\n'
        cell = cell.encode('utf8', 'replace')
        if args.bg:
            self.bg_processes.append(p)
            self._gc_bg_processes()
            if args.out:
                self.shell.user_ns[args.out] = p.stdout
            if args.err:
                self.shell.user_ns[args.err] = p.stderr
            self.job_manager.new(self._run_script, p, cell, daemon=True)
            if args.proc:
                self.shell.user_ns[args.proc] = p
            return

        try:
            out, err = p.communicate(cell)
        except KeyboardInterrupt:
            try:
                p.send_signal(signal.SIGINT)
                time.sleep(0.1)
                if p.poll() is not None:
                    print("Process is interrupted.")
                    return
                p.terminate()
                time.sleep(0.1)
                if p.poll() is not None:
                    print("Process is terminated.")
                    return
                p.kill()
                print("Process is killed.")
            except OSError:
                pass
            except Exception as e:
                print("Error while terminating subprocess (pid=%i): %s" \
                      % (p.pid, e))
            return
        out = py3compat.bytes_to_str(out)
        err = py3compat.bytes_to_str(err)
        if args.out:
            self.shell.user_ns[args.out] = out
        else:
            sys.stdout.write(out)
            sys.stdout.flush()
        if args.err:
            self.shell.user_ns[args.err] = err
        else:
            sys.stderr.write(err)
            sys.stderr.flush()

def get_ipython():
    return ScriptIPython(caller_locals())



import unittest
from csv import DictReader
from metapack import MetapackDoc


from metapack import MetapackPackageUrl,  MetapackUrl, ResourceError, Downloader
from metapack.cli.core import (make_filesystem_package, make_s3_package, make_excel_package, make_zip_package, make_csv_package,
                                PACKAGE_PREFIX, cli_init )
from rowgenerators import get_generator, RowGeneratorError, parse_app_url
from metatab.generate import TextRowGenerator
from metapack.test.support import test_data

downloader = Downloader()



class TestPublish(unittest.TestCase):


    def test_config(self):
        from metapack.cli.core import get_config

        print(get_config())


    @unittest.skip("Need Sensible Credentials")
    def test_publish_wp(self):
        pass

        from wordpress_xmlrpc import Client, WordPressPost
        from wordpress_xmlrpc.methods.posts import GetPosts, NewPost
        from wordpress_xmlrpc.methods.users import GetUserInfo

        wp = Client('url', 'user', 'password')

        print(wp.call(GetPosts()))

        print(wp.call(GetUserInfo()))

        post = WordPressPost()
        post.title = 'My new title'
        post.content = 'This is the body of my new post.'
        post.terms_names = {
          'post_tag': ['test', 'firstpost'],
          'category': ['Introductions', 'Tests']
        }
        post.custom_fields = [
            {'key': 'cf_1','value':'cfv_1'},
            {'key': 'cf_2', 'value': 'cfv_2'}
        ]
        wp.call(NewPost(post))

    @unittest.skip("Need Sensible Credentials")
    def test_list_wp(self):

        from wordpress_xmlrpc import Client, WordPressPost
        from wordpress_xmlrpc.methods.posts import GetPosts
        from wordpress_xmlrpc.methods.taxonomies import GetTaxonomies
        from wordpress_xmlrpc.methods.media import GetMediaLibrary

        wp = Client('url', 'user', 'password')

        for p in wp.call(GetPosts()):
            print(p, p.slug,)
            for t in  p.custom_fields:
                print ('   ',t)
            print('   ','---')
            for img in wp.call(GetMediaLibrary(dict(parent_id=p.id))):
                print('   ',img.id, img.metadata, img.link )


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

"""
CLI program for managing packages
"""

import sys
from os import getenv
from metapack import Downloader
from metapack.cli.core import prt, err, MetapackCliMemo, cli_init
from metapack.appurl import SearchUrl
from tabulate import tabulate
import shutil
from itertools import islice
from bisect import bisect
from terminaltables import AsciiTable, SingleTable, DoubleTable, GithubFlavoredMarkdownTable

downloader = Downloader()


def run(subparsers):

    parser = subparsers.add_parser(
        'run',
        help='Generate rows for a resource or reference ',
        epilog='Cache dir: {}\n'.format(str(downloader.cache.getsyspath('/'))))

    parser.set_defaults(run_command=run_run)

    #
    # What resource to run

    reference_group = parser.add_argument_group("Resource or Reference specification")

    group = reference_group.add_mutually_exclusive_group()

    group.add_argument('-r', '--resource',
                        help="Name of the Root.Resource to run  ")

    group.add_argument('-f', '--reference',
                        help="Name of the Root.Reference to run  ")



    #
    # Output format

    format_group = parser.add_argument_group("Output format")

    group = format_group.add_mutually_exclusive_group()

    group.add_argument('-c', '--CSV', default=False, action='store_true',
                             help='Output as CSV')

    group.add_argument('-t', '--tabs', default=False, action='store_true',
                       help='Output as tab-delimited rows')

    group.add_argument('-j', '--json', default=False, action='store_true',
                       help='Output JSON lines. Each line is a completel JSON object')

    group.add_argument('-y', '--yaml', default=False, action='store_true',
                       help='Output YAML')

    group.add_argument('-T', '--table', default=False, action='store_true',
                       help='Output 20 rows in a table format. Truncates columns to width of terminal')

    #
    # Table format options

    table_group = parser.add_argument_group('Options for Table ( -T ) format')

    table_group.add_argument('-p', '--pivot', default=False, action='store_true',
                       help='When outputting a table, transpose rows for columns')

    table_group.add_argument('-m', '--markdown', default=False, action='store_true',
                       help='When outputting a table, use Markdown format')

    table_group.add_argument('metatabfile', nargs='?',
                        help="Path or URL to a metatab file. If not provided, defaults to 'metadata.csv' ")

    #
    # Other options

    output_group = parser.add_argument_group("General output options")

    output_group.add_argument('-S', '--sample', type=str, help="Sample values from a column")

    output_group.add_argument('-L', '--limit', type=int,
                       help="Limit the number of output rows ")

    parser.set_defaults(handler=None)

def list_rr(doc):

    d = []
    for r in doc.resources():
        d.append(('Resource', r.name, r.url))

    for r in doc.references():
        d.append(('Reference', r.name, r.url))

    prt(tabulate(d, 'Type Name Url'.split()))


def get_resource(m):


    if m.resource:
        r = m.doc.resource(m.resource)
        return r if r else m.doc.reference(m.resource)
    elif m.args.resource:
        return m.doc.resource(m.args.resource)
    elif m.args.reference:
        return m.doc.reference(m.args.reference)
    else:
        err("Must specify a resource or a reference")


def run_run(args):

    m = MetapackCliMemo(args, downloader)

    r = get_resource(m)

    doc = m.doc

    if not r:
        list_rr(doc)
        sys.exit(0)


    if not r:
        prt("ERROR: No resource or reference for '{}' valid terms are:\n".format(m.args.resource))
        list_rr(doc)
        sys.exit(1)

    if m.args.sample:

        from collections import Counter

        limit = args.limit if args.limit else 5000

        c = Counter( r[m.args.sample] for r in islice(r.iterrows, None, limit))

        prt(tabulate(c.most_common(10), headers='Value Count'.split()))


    elif m.args.table:

        limit = args.limit if args.limit else 20

        t_width = shutil.get_terminal_size()[0]

        rows = list(islice(r, None, limit))

        if m.args.pivot:
            rows = list(zip(*rows))
            header = ['Column Name'] + ['Column{}'.format(i) for i in range(1,len(rows[1])) ]
            rows = [header] + rows

        # Only display the colums that will fit in the terminal window
        # If I were good, this would be a comprehension
        max_lengths = [0] * len(rows[0])
        for row in rows:
            for i, col in enumerate(row):
                v = str(col)
                if '\n' in v:
                    ln = max(len(l)+3 for l in v.splitlines() )
                else:
                    ln = len(str(col))+3

                max_lengths[i] = max(max_lengths[i], ln )

        display_cols = bisect([ sum(max_lengths[:i]) for i in range(1,len(max_lengths)+1)],t_width-4)

        # print(tabulate( [row[:display_cols] for row in rows], tablefmt="psql"))

        if m.args.markdown:
            table_class = GithubFlavoredMarkdownTable
        else:
            table_class = SingleTable

        at = table_class([row[:display_cols] for row in rows])
        at.inner_row_border = True
        print (at.table)

    else:
        import csv

        if args.limit:
            gen_wrap = lambda g: islice(g, None, args.limit)
        else:
            gen_wrap = lambda g: g

        if m.args.json:
            for i,j in enumerate(gen_wrap(r.iterjson())):
                print(j)

        elif m.args.yaml:
            for i,j in enumerate(gen_wrap(r.iteryaml())):
                print(j, j,end='')
        elif m.args.tabs:
            w = csv.writer(sys.stdout, delimiter='\t')
            for i,j in enumerate(gen_wrap(r)):
                w.writerow(j)

        else: # m.args.csv:
            w = csv.writer(sys.stdout)
            for i,j in enumerate(gen_wrap(r)):
                w.writerow(j)




# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
Root program for metapack programs

"""

from pkg_resources import iter_entry_points, get_distribution, DistributionNotFound
from .core import prt
import argparse

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
   # package is not installed
   pass

from metapack.cli.core import  cli_init


def mp():
    cli_init()

    parser = argparse.ArgumentParser(
        prog='metapack',
        description='Create and manipulate metatab data packages. ')

    parser.add_argument('-v', '--version', default=False, action='store_true',
                             help='Print Metapack versions')

    parser.add_argument('-V', '--versions', default=False, action='store_true',
                        help='Print version of several important packages')

    parser.add_argument('-C', '--cache', default=False, action='store_true',
                        help='Print the location of the cache')

    subparsers = parser.add_subparsers(help='Commands')

    for ep in iter_entry_points(group='mt.subcommands'):

        f = ep.load()
        f(subparsers)

    args = parser.parse_args()

    if args.version:
        prt(get_distribution('metapack'))

    if args.cache:
        from shlex import quote
        from metapack import MetapackDoc, Downloader
        downloader = Downloader()

        prt(quote(downloader.cache.getsyspath('/')))

    elif args.versions:

        from pkg_resources import EntryPoint

        prt('--- Main Packages')

        main_packages = ('metapack', 'metatab','rowgenerators','publicdata')

        for pkg_name in main_packages:
            try:
                prt(get_distribution(pkg_name))
            except (DistributionNotFound, ModuleNotFoundError) as e:
                # package is not installed

                pass

        prt('')
        prt('--- Subcommands')

        for ep in iter_entry_points(group='mt.subcommands'):
            prt(ep.name,  ep.dist)

    else:

        args.run_command(args)


# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""Support for IPython and Python kernels in Jupyter Notebooks"""


from metapack import open_package as op
from os import getcwd
from metapack.util import walk_up
from os.path import getmtime, join, exists
from metapack.exc import PackageError
from rowgenerators import RowGeneratorError

def caller_locals():
    """Get the local variables in the caller's frame."""
    import inspect
    frame = inspect.currentframe()
    try:
        return frame.f_back.f_back.f_locals
    finally:
        del frame

def in_build():
    """Return True if running in a build, rather than interactively in Jupyter"""
    return 'metatab_doc' in caller_locals()


def open_source_package(dr=None):

    if dr is None:
        dr = getcwd()

    for i, e in enumerate(walk_up(dr)):

        if 'metadata.csv' in e[2]:
            return op(join(e[0], 'metadata.csv'))

        if i > 2:
            break

    return None

def open_package(locals=None, dr=None):
    """Try to open a package with the metatab_doc variable, which is set when a Notebook is run
    as a resource. If that does not exist, try the local _packages directory"""


    if locals is None:
        locals = caller_locals()

    try:
        # Running in a package build
        return op(locals['metatab_doc'])

    except KeyError:
        # Running interactively in Jupyter


        package_name = None
        build_package_dir = None
        source_package = None

        if dr is None:
            dr = getcwd()

        for i, e in enumerate(walk_up(dr)):

            if 'metadata.csv' in e[2]:
                source_package = join(e[0],'metadata.csv')
                p = op(source_package)
                package_name = p.find_first_value("Root.Name")

                if not package_name:
                    raise PackageError("Source package in {} does not have root.Name term".format(e[0]))

                if '_packages' in e[1]:
                    build_package_dir = join(e[0], '_packages')

                break

            if i > 2:
                break

        if build_package_dir and package_name and exists(join(build_package_dir, package_name)):
            # Open the previously built package
            built_package = join(build_package_dir, package_name)
            try:
                return op(built_package)
            except RowGeneratorError as e:
                pass # Probably could not open the metadata file.

        if source_package:
            # Open the source package
            return op(source_package)

    raise PackageError("Failed to find package, either in locals() or above dir '{}' ".format(dr))


def rebuild_schema(doc, r, df):
    """Rebuild the schema for a resource based on a dataframe"""
    import numpy as np

    # Re-get the resource in the doc, since it may be different.
    try:
        r = doc.resource(r.name)
    except AttributeError:
        # Maybe r is actually a resource name
        r = doc.resource(r)

    def alt_col_name(name, i):
        import re

        if not name:
            return 'col{}'.format(i)

        return re.sub('_+', '_', re.sub('[^\w_]', '_', str(name)).lower()).rstrip('_')

    df_types = {
        np.dtype('O'): 'text',
        np.dtype('int64'): 'integer',
        np.dtype('float64'): 'number'
    }

    try:
        df_index_frame = df.index.to_frame()
    except AttributeError:
        df_index_frame = None

    def get_col_dtype(c):

        c = str(c)

        try:
            return df_types[df[c].dtype]
        except KeyError:
            # Maybe it is in the index?
            pass

        try:
            return df_types[df_index_frame[c].dtype]
        except TypeError:
            # Maybe not a multi-index
            pass

        if c == 'id' or c == df.index.name:
            return df_types[df.index.dtype]

        return 'unknown'

    columns = []
    schema_term = r.schema_term[0]

    if schema_term:

        old_cols = { c['name'].value: c.properties for c in schema_term.children }
        for c in schema_term.children:
            schema_term.remove_child(c)

        schema_term.children = []

    else:
        old_cols = {}
        schema_term = doc['Schema'].new_term('Table', r.schema_name)

    index_names = [n if n else "id" for n in df.index.names]

    for i, col in enumerate(index_names + list(df.columns)):
        acn = alt_col_name(col, i) if alt_col_name(col, i) != str(col) else ''

        d = {'name': col, 'datatype': get_col_dtype(col), 'altname': acn}

        if col in old_cols.keys():
            lookup_name = col
        elif acn in old_cols.keys():
            lookup_name = acn
        else:
            lookup_name = None

        if lookup_name and lookup_name in old_cols:

            for k,v in schema_term.properties.items():

                old_col = old_cols.get(lookup_name)

                for k,v in old_col.items():
                    if k != 'name' and v :
                        d[k] = v

        columns.append(d)

    for c in columns:

        name = c['name']
        del c['name']
        datatype = c['datatype']
        del c['datatype']
        altname = c['altname']
        del c['altname']

        schema_term.new_child('Column', name, datatype=datatype, altname=altname, **c)


def rewrite_schema(r, df, doc = None):
    """Rebuild the schema for a resource based on a dataframe and re-write the doc"""

    from metatab.cli.core import write_doc

    if doc is None:
        doc = open_source_package()

    rebuild_schema(doc, r, df)

    write_doc(doc, doc.ref)

def interactive_rewrite_schema(r, df, doc = None):
    """Rebuild the schema for a resource based on a dataframe and re-write the doc,
    but only if running the notebook interactively, not while building"""

    from metatab.cli.core import write_doc

    if  'metatab_doc' in caller_locals():
        return False

    if doc is None:
        doc = open_source_package()

    rewrite_schema(r, df, doc)


    return True


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
Manage AWS S3 bucket permissions and users
"""

import metatab
import mimetypes
import sys
from os import getenv, getcwd
from os.path import join, basename

from metatab import _meta, DEFAULT_METATAB_FILE, resolve_package_metadata_url, MetatabDoc, open_package, MetatabError
from metatab.cli.core import err, metatab_info
from rowgenerators import  Url
from .core import prt, warn, err, write_doc
from .metasync import update_dist
import json

from os.path import dirname, join, abspath
import json

from os.path import dirname, join, abspath


def metaaws():
    import argparse

    parser = argparse.ArgumentParser(prog='metakan',
                                     description='CKAN management of Metatab packages')

    parser.add_argument('-p', '--profile_name', type=str, nargs='?', help='Name of boto/aws credentials file')
    parser.set_defaults(subcommand=None)
    asp = parser.add_subparsers(title='AWS commands', help='AWS commands')

    sp = asp.add_parser('list-buckets', help="List buckets")
    sp.set_defaults(subcommand=list_remotes)

    sp = asp.add_parser('list-users', help="List users")
    sp.set_defaults(subcommand=list_users)

    sp = asp.add_parser('list-bucket-users', help="List users with permissions on a bucket")
    sp.set_defaults(subcommand=list_bucket_users)
    sp.add_argument('bucket', help='Bucket name')

    sp = asp.add_parser('init-bucket', help="Initialize ambry buckets")
    sp.set_defaults(subcommand=init_bucket)
    sp.add_argument('bucket_name', help='Bucket name')

    sp = asp.add_parser('new-user', help="Create a new IAM user")
    sp.set_defaults(subcommand=new_user)
    sp.add_argument('user_name', help='User name')

    sp = asp.add_parser('delete-user', help="Delete an IAM user")
    sp.set_defaults(subcommand=delete_user)
    sp.add_argument('user_name', help='User name')

    sp = asp.add_parser('perm', help="Add a permission to a user on a bucket")
    sp.set_defaults(subcommand=perm)
    sp.add_argument('-w', '--write', default=False, action='store_true', help="Also add write permission")
    sp.add_argument('-d', '--delete', default=False, action='store_true', help="Remove the permission instead")
    sp.add_argument('user_name', help='User name')
    sp.add_argument('bucket', help='Bucket name, possibly with prefix')

    sp = asp.add_parser('test-user', help="Test user access")
    sp.set_defaults(subcommand=test_user)
    sp.add_argument('user_name', help='User name')
    sp.add_argument('bucket', help='Bucket name')

    args = parser.parse_args()

    if args.subcommand is None:
        parser.print_help()
    else:
        args.subcommand(args)  # Note the calls to sp.set_defaults(subcommand=...)


USER_PATH = '/metatab/'
TOP_LEVEL_DIRS = ( 'public', 'restricted', 'private')

policy_name = 'ambry-s3'  # User policy name


def get_client(cli_args, service, *args, **kwargs):
    import boto3
    session = boto3.Session(profile_name=cli_args.profile_name)

    return session.client(service, *args, **kwargs)


def get_resource(cli_args, service, *args, **kwargs):
    import boto3
    session = boto3.Session(profile_name=cli_args.profile_name)

    return session.resource(service, *args, **kwargs)


def list_remotes(args):
    client = get_client(args, 's3')

    r = client.list_buckets()
    for e in r['Buckets']:
        print(e['Name'])


def new_user(args):
    from botocore.exceptions import ClientError

    client = get_client(args, 'iam')

    try:
        r = client.get_user(UserName=args.user_name)
        prt("User already exists")

    except ClientError:
        r = client.create_user(Path=USER_PATH, UserName=args.user_name)

        iam = get_resource(args, 'iam')

        user_name = r['User']['UserName']

        user = iam.User(user_name)

        key_pair = user.create_access_key_pair()

        prt("Created user : {}".format(user.user_name))
        prt("arn          : {}".format(user.arn))
        prt("Access Key   : {}".format(key_pair.id))
        prt("Secret Key   : {}".format(key_pair.secret))


def make_group_policy(bucket, prefix, write=False):
    policy_path = join(dirname(abspath(metatab.__file__)), 'support', 'group-policy.json')

    with open(policy_path) as f:
        policy_doc = f.read()

    doc = json.loads(policy_doc)

    arn = 'arn:aws:s3:::'

    def get_statement(name):
        for s in doc['Statement']:
            if s['Sid'] == name:
                return s

    def del_statement(name):
        for i, s in enumerate(doc['Statement']):
            if s['Sid'] == name:
                del doc['Statement'][i]

    get_statement('bucket')['Resource'].append(arn + bucket)
    get_statement('read')['Resource'].append(arn + bucket + '/' + prefix.strip('/') + '/*')

    if write:
        get_statement('write')['Resource'].append(arn + bucket + '/' + prefix.strip('/') + '/*')
    else:
        del_statement('write')

    return json.dumps(doc)


arn_prefix = 'arn:aws:s3:::'


def user_dict_to_policy(d):
    policy_path = join(dirname(abspath(metatab.__file__)), 'support', 'user-policy.json')

    with open(policy_path) as f:
        policy_doc = f.read()

    doc = json.loads(policy_doc)

    buckets = set()
    reads = set()
    writes = set()

    for (bucket, prefix), mode in d.items():

        buckets.add(bucket)

        if mode == 'R':
            reads.add((bucket, prefix))
        elif mode == 'W':
            reads.add((bucket, prefix))
            writes.add((bucket, prefix))

    # Get a reference to a read, write or bucket statement
    def get_statement(name):
        for s in doc['Statement']:
            if s['Sid'] == name:
                return s

    get_statement('list')['Resource'] = [arn_prefix + b for b in buckets]
    get_statement('bucket')['Resource'] = [arn_prefix + b for b in buckets]
    get_statement('read')['Resource'] = ["{}{}/{}/*".format(arn_prefix, b, p) for b, p in reads]

    if writes:
        get_statement('write')['Resource'] = ["{}{}/{}/*".format(arn_prefix, b, p) for b, p in writes]
    else:
        # Can't have a section with no resources
        doc['Statement'] = [s for s in doc['Statement'] if s['Sid'] != 'write']

    # Add permission to list buckets with read access
    from collections import defaultdict
    prefixes_per_bucket = defaultdict(set)
    for (bucket, prefix) in reads:
        prefixes_per_bucket[bucket].add(prefix)

    for bucket, prefixes in prefixes_per_bucket.items():
        doc['Statement'].append(
            {
                "Sid": "List{}".format(''.join(p.title() for p in bucket.split('.'))),
                "Action": ["s3:ListBucket"],
                "Effect": "Allow",
                "Resource": ["{}{}".format(arn_prefix, bucket)],
                "Condition": {"StringLike": {"s3:prefix": ["{}/*".format(prefix) for prefix in prefixes]}}
            }
        )

    return json.dumps(doc, indent=4)


def user_policy_to_dict(doc):
    """Convert a bucket policy to a dict mapping principal/prefix names to 'R' or 'W' """

    import json

    if not isinstance(doc, dict):
        doc = json.loads(doc)

    d = {}

    def get_statement(name):
        for s in doc['Statement']:
            if s['Sid'] == name:
                return s

    for r in get_statement('read')['Principal']['AWS']:
        bucket, prefix = r.replace(arn_prefix, '').replace('/*', '').split('/')
        d[(bucket, prefix.strip('/'))] = 'R'

    try:
        for r in get_statement('write')['Resource']:
            bucket, prefix = r.replace(arn_prefix, '').replace('/*', '').split('/')
            d[(bucket, prefix.strip('/'))] = 'W'
    except TypeError:
        pass  # No write section

    return d


def make_bucket_policy_statements(bucket):
    """Return the statemtns in a bucket policy as a dict of dicts"""
    import yaml
    from os.path import dirname, join, abspath
    import copy
    import metatab

    with open(join(dirname(abspath(metatab.__file__)), 'support', 'policy_parts.yaml')) as f:
        parts = yaml.load(f)

    statements = {}

    cl = copy.deepcopy(parts['list'])
    cl['Resource'] = arn_prefix + bucket
    statements['list'] = cl

    cl = copy.deepcopy(parts['bucket'])
    cl['Resource'] = arn_prefix + bucket
    statements['bucket'] = cl

    for sd in TOP_LEVEL_DIRS:
        cl = copy.deepcopy(parts['read'])
        cl['Resource'] = arn_prefix + bucket + '/' + sd + '/*'
        cl['Sid'] = cl['Sid'].title() + sd.title()

        statements[cl['Sid']] = cl

        cl = copy.deepcopy(parts['write'])
        cl['Resource'] = arn_prefix + bucket + '/' + sd + '/*'
        cl['Sid'] = cl['Sid'].title() + sd.title()

        statements[cl['Sid']] = cl

        cl = copy.deepcopy(parts['listb'])
        cl['Resource'] = arn_prefix + bucket
        cl['Sid'] = cl['Sid'].title() + sd.title()
        cl['Condition']['StringLike']['s3:prefix'] = [sd + '/*']

        statements[cl['Sid']] = cl

    return statements


def bucket_dict_to_policy(args, bucket_name, d):
    """
    Create a bucket policy document from a permissions dict.

    The dictionary d maps (user, prefix) to 'R' or 'W'.

    :param bucket_name:
    :param d:
    :return:
    """

    import json

    iam = get_resource(args, 'iam')

    statements = make_bucket_policy_statements(bucket_name)

    user_stats = set()  # statement tripples

    for (user, prefix), mode in d.items():

        user_stats.add((user, 'list'))
        user_stats.add((user, 'bucket'))

        if mode == 'R':
            user_stats.add((user, 'Read' + prefix.title()))
            user_stats.add((user, 'List' + prefix.title()))
        elif mode == 'W':
            user_stats.add((user, 'List' + prefix.title()))
            user_stats.add((user, 'Read' + prefix.title()))
            user_stats.add((user, 'Write' + prefix.title()))

    users_arns = {}

    for user_name, section in user_stats:
        section = statements[section]

        if user_name not in users_arns:
            user = iam.User(user_name)
            users_arns[user.name] = user
        else:
            user = users_arns[user_name]

        section['Principal']['AWS'].append(user.arn)

    for sid in list(statements.keys()):
        if not statements[sid]['Principal']['AWS']:
            del statements[sid]

    return json.dumps(dict(Version="2012-10-17", Statement=list(statements.values())), indent=4)


def bucket_policy_to_dict(policy):
    """Produce a dictionary of read, write permissions for an existing bucket policy document"""
    import json

    if not isinstance(policy, dict):
        policy = json.loads(policy)

    statements = {s['Sid']: s for s in policy['Statement']}

    d = {}

    for rw in ('Read', 'Write'):
        for prefix in TOP_LEVEL_DIRS:
            sid = rw.title() + prefix.title()

            if sid in statements:

                if isinstance(statements[sid]['Principal']['AWS'], list):

                    for principal in statements[sid]['Principal']['AWS']:
                        user_name = principal.split('/').pop()
                        d[(user_name, prefix)] = rw[0]
                else:
                    user_name = statements[sid]['Principal']['AWS'].split('/').pop()
                    d[(user_name, prefix)] = rw[0]

    return d


def delete_user(args):
    from botocore.exceptions import ClientError

    client = get_client(args, 'iam')

    try:
        resource = get_resource(args, 'iam')
        user = resource.User(args.user_name)

        for key in user.access_keys.all():
            prt("Deleting user key: {}".format(key))
            key.delete()

        for policy in user.policies.all():
            prt("Deleting user policy: {}".format(policy.name))
            policy.delete()

        response = client.delete_user(UserName=args.user_name)
        prt("Deleted user: {}".format(args.user_name))

    except ClientError as e:
        err("Could not delete user: {}".format(e))


def init_bucket(args):
    from botocore.exceptions import ClientError
    import json

    s3 = get_resource(args, 's3')

    b = s3.Bucket(args.bucket_name)

    b.create()


def split_bucket_name(bucket, default='public'):
    if '/' in bucket:
        bn, prefix = bucket.split('/', 1)
    elif default != False:
        bn = bucket
        prefix = default
    else:
        bn = bucket
        prefix = None

    return bn, prefix


def perm(args):
    from botocore.exceptions import ClientError
    import json

    iam = get_resource(args, 'iam')

    bn, prefix = split_bucket_name(args.bucket, default=False)

    if not prefix:
        prefixes = TOP_LEVEL_DIRS
    else:
        prefixes = [prefix]

    user = iam.User(args.user_name)

    b = get_resource(args, 's3').Bucket(bn)

    try:

        bucket_policy = b.Policy().policy
        perms = bucket_policy_to_dict(bucket_policy)

    except ClientError:
        perms = {}
        bucket_policy = None

    for prefix in prefixes:
        if args.delete:
            if (user.arn, prefix) in perms:
                del perms[(user.name, prefix)]
                prt("Removed {}/{} from {}".format(bn, prefix, user.name))
        else:

            if args.write:
                perms[(user.name, prefix)] = 'W'
                prt("Added write {}/{} to {}".format(bn, prefix, user.name))
                has_writes = True
            else:
                perms[(user.name, prefix)] = 'R'
                prt("Added read {}/{} to {}".format(bn, prefix, user.name))

    if perms:
        b = get_resource(args, 's3').Bucket(bn)
        policy = bucket_dict_to_policy(args, bn, perms)

        try:
            b.Policy().put(Policy=policy)
        except Exception as e:
            print(policy)
            raise



    elif bucket_policy:
        bucket_policy.delete()


def list_users(args):
    from botocore.exceptions import ClientError
    import tabulate

    client = get_client(args, 'iam')
    iam = get_resource(args, 'iam')

    records = []
    headers = 'Name access ARN'.split()

    users = client.list_users(PathPrefix='/')

    for user_info in users['Users']:
        user = iam.User(user_info['UserName'])

        records.append([user.name, list(user.access_keys.all())[0].id, user.arn])

    print(tabulate.tabulate(records, headers))


def list_bucket_users(args):
    from botocore.exceptions import ClientError
    import tabulate

    client = get_client(args, 'iam')
    iam = get_resource(args, 'iam')

    b = get_resource(args, 's3').Bucket(args.bucket)

    try:
        perms = bucket_policy_to_dict(b.Policy().policy)
    except ClientError:
        perms = {}

    records = []
    headers = ['Name'] + list(TOP_LEVEL_DIRS)

    users = client.list_users(PathPrefix='/')

    for user_info in users['Users']:

        user = iam.User(user_info['UserName'])

        row = [user.name]

        perm_count = 0
        for prefix in TOP_LEVEL_DIRS:
            perm = perms.get((user.name, prefix))
            perm_count += int(perm is not None)
            row.append(perm)

        if perm_count:
            records.append(row)

    print(tabulate.tabulate(records, headers))


def get_iam_account(l, args, user_name):
    """Return the local Account for a user name, by fetching User and looking up
    the arn. """

    iam = get_resource(args, 'iam')
    user = iam.User(user_name)
    user.load()

    return l.find_or_new_account(user.arn)


def test_user(args):
    from botocore.exceptions import ClientError
    import boto3

    account = get_iam_account(l, args, args.user_name)

    if not account.access_key:
        err("Can't test user {}; library does not have record for account ( by arn ) ".format(args.user_name))

    session = boto3.Session(aws_access_key_id=account.access_key,
                            aws_secret_access_key=account.secret)

    root_s3 = get_resource(args, 's3')
    s3 = session.resource('s3')

    bn, prefix = split_bucket_name(args.bucket, default=None)

    root_bucket = root_s3.Bucket(bn)
    bucket = s3.Bucket(bn)

    prefixes = [prefix] if prefix else TOP_LEVEL_DIRS

    for prefix in prefixes:
        k = prefix + '/test/' + args.user_name
        rk = k + '-root'

        ro = root_bucket.put_object(Key=rk, Body=args.user_name)

        try:
            o = bucket.Object(rk)
            c = o.get()
            read = True
        except ClientError as e:
            read = False

        try:
            o = bucket.put_object(Key=k, Body=args.user_name)
            write = True
        except ClientError as e:
            write = False

        try:
            o.delete()
            delete = True
        except ClientError as e:
            delete = False

        # ro.delete()

        prt("{:<35s} {:<5s} {:<5s} {:<6s} {}".format(k, 'read' if read else '',
                                                     'write' if write else '',
                                                     'delete' if delete else '',
                                                     'no access' if not any((read, write, delete)) else ''))
